#!/usr/bin/env python3
"""PiOLED display service and driver
"""

import importlib.metadata
__version__ = importlib.metadata.version(__package__ or __name__)

#==========================================================
#
#  Chris Nelson, Copyright 2024-2026
#
#==========================================================

import time
import datetime
import argparse
import sys
import os
import signal
import ast
from pathlib import Path
from types import SimpleNamespace
from threading import Thread
from importlib_resources import files as ir_files

from luma.core import cmdline
from luma.core.render import canvas
from PIL import ImageFont

from cjnfuncs.core          import set_toolname, logging, set_logging_level, periodic_log, setuplogging
from cjnfuncs.configman     import config_item
from cjnfuncs.resourcelock  import resource_lock
from cjnfuncs.rwt           import run_with_timeout
from cjnfuncs.deployfiles   import deploy_files
import cjnfuncs.core as core


# Configs / Constants
TOOLNAME =                      'PiOLED'
CONFIG_FILE =                   'PiOLED_server.cfg'       # Abs, or relative to the core.tool.config_dir
PIOLED_GO_FLAG =                'PiOLED_go_flag'
PIOLED_FILE_LOCK =              'PiOLED_file_lock'

SERVER_FILE_LOGGING_FORMAT =    '{asctime} {module:>15}.{funcName:20} * {levelname:>8}:  {message}'
SERVER_CONS_LOGGING_FORMAT =              '{module:>15}.{funcName:20} * {levelname:>8}:  {message}'
CLI_CONSOLE_LOGGING_FORMAT =              '{module:>15}.{funcName:20} - {levelname:>8}:  {message}'

DISPLAY_DRIVER_FONT =           'C&C Red Alert [INET].ttf'  # Default font for _oneliner    # TODO configurable, size and color
DISPLAY_DRIVER_SIZE =           12
DISPLAY_DRIVER_COLOR =          'white'
PIOLED_TH_EXIT =                -99
PIOLED_TH_PAUSE =               -98
PIOLED_NEWMSG =                 -1
PIOLED_SAVE =                   -2
PIOLED_RESTORE =                -3

PIOLED_PAGE_TIME =              7
PIOLED_INTER_PAGE_TIME =        0.3
PIOLED_INTER_MESSAGE_SET_TIME = 1

pioled_logger = logging.getLogger('cjn_PiTools.PiOLED')


set_toolname (TOOLNAME)
# print (core.tool)

pioled_go_flag =                resource_lock(PIOLED_GO_FLAG)
pioled_file_lock =              resource_lock(PIOLED_FILE_LOCK)



#=====================================================================================
#=====================================================================================
#   c l i e n t - s i d e   p i o l e d _ d i s p l a y _ d r i v e r
#=====================================================================================
#=====================================================================================

class pioled_display_driver:
    """Driver for sending messages on the OLED display via a display_file
    This driver/handler is instantiated within the client/tool script/app.
    The config file is not used (used only by the server).
    """
    def __init__(self, queue, display_file,
                 name='pioled_driver',
                 page_time=PIOLED_PAGE_TIME,
                 inter_page_time=PIOLED_INTER_PAGE_TIME,
                 inter_message_set_time=PIOLED_INTER_MESSAGE_SET_TIME
                 ):
        global pioled_logger
        self.queue = queue
        self.display_file = Path(display_file)
        self.name = name
        self.page_time = page_time
        self.inter_page_time = inter_page_time
        self.inter_message_set_time = inter_message_set_time
        self.saved_message_set = {'cmd:':None, 'cnt':None, 'pages':[[{'x':0,  'y':0,  'size':18, 'text':"Warning"},
                                                                     {'x':10, 'y':20, 'size':12, 'text':"Message set Restored"},
                                                                     {'x':10, 'y':32, 'size':12, 'text':"before Saved"}]]}

        # Check access to pioled_file_lock and display_file
        if not pioled_file_lock.get_lock(lock_info='pioled_display_driver startup file access check', timeout=2):
            pioled_logger.warning (f"Failed to get pioled_file_lock during pioled_display_driver startup - Continuing. Current lock owner: <{pioled_file_lock.get_lock_info()}>")
        else:
            try:
                self.display_file.touch()
                self.display_file.unlink()
            except Exception as e:
                pioled_logger.warning (f"Unable to access oled display_file <{self.display_file}> - Continuing\n  {type(e).__name__}: {e}")
            pioled_file_lock.unget_lock(where_called="pioled_display_driver startup file access check")


    def start(self):
        self.this_thread = Thread(target=self.message_loop, name=self.name, daemon=True)    # daemon so that if the main thread errors this thread is auto-killed, avoiding hang
        self.this_thread.start()
        pioled_logger.debug (f"pioled message_loop thread created using display_file <{self.display_file}>")
        return self.this_thread


    def blank(self):
        pioled_logger.debug ("pioled_blank")
        self.message_page([[0, 0, 12, '']])     # Not sent thru the queue as the blank would be saved and restored
        time.sleep (0.01)                       # Allow time for blank to be processed


    def oneliner(self, x, y, size, text, message_time=0, font=None, color=None, cmd=None):
        # message_time is blocking to the main code.  Alternately send thru the queue with a second blank page.
        pioled_logger.debug (f"pioled_message - <{text}>")
        self.queue.put ({'cmd':cmd, 'pages':[[{'x':x, 'y':y, 'size':size, 'text':text, 'font':font, 'color':color}]]})
        if message_time:
            time.sleep (message_time)
            self.blank()
        else:
            time.sleep (0.01)                   # Allow time for queue to be processed


    def message_loop(self):
        """Run this functions in a thread, sending a full set of message pages thru the queue
        per the `cmd` and `cnt` settings.

        New message is displayed in < 0.15 sec
        """

        message_set = {'cmd:':None, 'cnt':None, 'pages':[]}
        while True:                                                 # "Top-level Loop While"
            prior_message_set = message_set
            message_set = self.queue.get()
            pioled_logger.debug (f"Received message_set: <{message_set}>")

            try:
                cmd = message_set.get('cmd', PIOLED_NEWMSG)
                if cmd == PIOLED_SAVE:
                    self.saved_message_set = prior_message_set
                    pioled_logger.debug ("Prior pioled message_set saved")
                if cmd == PIOLED_RESTORE:
                    message_set = self.saved_message_set
                    pioled_logger.debug ("Prior pioled message_set restored")

                cnt = message_set.get('cnt', -1)                    # Default loop endlessly
                pages = message_set.get('pages', None)

                if cmd == PIOLED_TH_EXIT:
                    if pages is not None:
                        self.message_page(pages[0])
                    else:
                        self.blank()
                    time.sleep (0.01)                               # Allow time for queue to be processed
                    pioled_logger.debug ("pioled message_loop() exiting")
                    sys.exit()                                      # On exit the locks are still set, cleared by the server
                
                elif cmd == PIOLED_TH_PAUSE:
                    if pages is not None:
                        self.message_page(pages[0])
                    else:
                        self.blank()
                    pioled_logger.debug ("pioled message_loop() paused")

                elif pages is None:
                    pioled_logger.debug ("Empty pages - blanking display")
                    self.blank()

                else:                                               # PIOLED_NEWMSG, PIOLED_SAVE, PIOLED_RESTORE
                    page_time =                 message_set.get('page_time', self.page_time)
                    inter_page_time =           message_set.get('inter_page_time', self.inter_page_time)
                    inter_message_set_time =    message_set.get('inter_message_set_time', self.inter_message_set_time)
                    pioled_logger.debug (f"Timings:  page_time={page_time}, inter_page_time={inter_page_time}, inter_message_set_time={inter_message_set_time}")
                    len_pages = len(pages)
                    while cnt != 0:                                 # "Iterate Pages While"
                        for index in range(len_pages):              # "For Pages"
                            self.message_page(pages[index])
                            wait_until = datetime.datetime.now() + datetime.timedelta(seconds=page_time)
                            while datetime.datetime.now() < wait_until:
                                if not self.queue.empty():          # Break wait if new message in queue
                                    break
                                time.sleep (0.1)

                            if not self.queue.empty():              # Break "For Pages" loop if new message in queue
                                break

                            if len_pages == 1:                      # If just single page, then leave it up and break out of "For Pages"
                                break


                            # Blank between individual pages or between full pages list repeat
                            wait_time = inter_page_time  if index < len_pages -1  else  inter_message_set_time
                            if wait_time > 0:                       # NOTE - inter_message_set_time = 0 leaves last page on display
                                self.blank()
                                wait_until = datetime.datetime.now() + datetime.timedelta(seconds=wait_time)
                                while datetime.datetime.now() < wait_until:
                                    if not self.queue.empty():      # Break wait if new message in queue
                                        break
                                    time.sleep (0.1)

                            if not self.queue.empty():              # Break "For Pages" loop if new message in queue
                                break

                        else:                                       # Code run on non-break "For Pages" exit
                            if cnt != -1:
                                cnt -= 1
                            continue                                # Continue "Iterate Pages While"

                        self.queue.task_done()                      # Code run on "For Pages" break
                        break                                       # Break "Iterate Pages While"
            except Exception as e:
                pioled_logger.warning (f"Failed to parse/display pioled message <{message_set}> - Skipping\n  {type(e).__name__}: {e}")
                self.queue.task_done()


    def message_page(self, message_list, message_time=0):
        # message_time is blocking to the main code if not sent thru a queue to pioled message_loop
        # returns 0 on success, 1 on failure
        xx = ""
        for line in message_list:
            if isinstance(line, list):
                xx += f"{{'x':{line[0]}, 'y':{line[1]}, 'size':{line[2]}, 'text':'''{line[3]}'''}}\n"
            elif isinstance(line, dict):
                xx += str(line) + '\n'
            else:
                pioled_logger.warning (f"Expecting list or dict, found type {type(line)}: <{line}> - Skipping\n  {message_list}")
                return 1

        if not pioled_file_lock.get_lock(lock_info='message_page'):
            periodic_log(f"Failed to get pioled_file_lock. Current lock owner: <{pioled_file_lock.get_lock_info()}> - Skipping",
                         category='get_lock', logger_name='pioled_logger', log_interval='1h', log_level=logging.WARNING)
            return 1

        pioled_logger.debug (f"Writing to display_file <{self.display_file}>:\n{xx[:-1]}")      # trim off final \n

        def _write_display_file():
            with self.display_file.open('wt') as ofile:
                ofile.write(xx)
                ofile.flush()

        try:
            run_with_timeout(_write_display_file, rwt_timeout=0.2)
        except Exception as e:
            periodic_log(f"Failed to write the display_file - Skipping\n  {type(e).__name__}: {e}",
                         category='display_file write', logger_name='pioled_logger', log_interval='1h', log_level=logging.WARNING)
            pioled_file_lock.unget_lock(where_called='message_page failed to write the display file')
            return 1

        pioled_go_flag.get_lock(lock_info='message_page')

        if message_time > 0:
            time.sleep(message_time)
            self.blank()

        return 0




#=====================================================================================
#=====================================================================================
#   S E R V E R
#=====================================================================================
#=====================================================================================

def service():
    """ Display content of DISPLAY_FILE
    """

    global pioled_disp        # Used in _oneliner() service_int_handler()
    signal.signal(signal.SIGINT,  service_int_handler)      # Ctrl-C
    signal.signal(signal.SIGTERM, service_int_handler)      # kill <pid>

    pioled_disp = pioled()
    _oneliner("PiOLED service started")
    time.sleep(2)
    _oneliner("")
    logging.debug (pioled_disp)

    pioled_file_lock.unget_lock(where_called="PiOLED service startup", force=True)  # Restarting the server will force unget the locks
    pioled_go_flag.unget_lock(where_called="PiOLED service startup", force=True)


    while 1:
        if pioled_go_flag.is_locked():
            if not display_file.exists():
                logging.warning(f"display_file <{display_file}> not found - Skipping")
                pioled_go_flag.unget_lock(where_called="service loop couldn't find the display_file", force=True)
                pioled_file_lock.unget_lock(where_called="service loop couldn't find the display_file", force=True)
            else:
                try:
                    contents = display_file.read_text()[:-1]            # drop final \n
                    logging.debug (f"PiOLED Server received at {datetime.datetime.now()}\n{contents}")

                    with canvas(pioled_disp.device) as draw:
                        for line in contents.split('\n'):
                            _line = ast.literal_eval(line)              # convert text to dict
                            x =     _line['x']
                            y =     _line['y']
                            size =  _line['size']
                            text =  _line['text']

                            try:
                                font_size = known_fonts.get_font(_line['font'], size)
                            except:
                                font_size = known_fonts.get_font(config.getcfg('Default_font', DISPLAY_DRIVER_FONT), size)  # TODO rework defaults
                            color = _line['color']  if ('color' in _line  and  _line['color'] is not None)  else config.getcfg('Default_color', DISPLAY_DRIVER_COLOR)

                            draw.text((x, y), text, font=font_size, fill=color)

                except Exception as e:
                    logging.warning (f"Error processing contents: {contents}\n  {type(e).__name__}: {e}")

                pioled_go_flag.unget_lock(where_called='service loop end', force=True)
                pioled_file_lock.unget_lock(where_called='service loop end', force=True)

        time.sleep(0.01)


def _oneliner(text, color='white'):
    """Private version used in service and service_int_handler"""
    logging.debug (f"<{text}>")
    with canvas(pioled_disp.device) as draw:
        draw.text((0, 0), text, font=known_fonts.get_font(DISPLAY_DRIVER_FONT, 12), fill=color)


def service_int_handler(signal, frame):
    logging.warning(f"Signal {signal} received.  Exiting.")
    _oneliner("PiOLED service exiting")
    time.sleep(2)
    pioled_file_lock.unget_lock(where_called='service_int_handler', force=True)
    pioled_go_flag.unget_lock(where_called='service_int_handler', force=True)
    sys.exit(0)     # Display is cleared by luma when the service process terminates


#=====================================================================================
#=====================================================================================
#   c l a s s   p i o l e d
#=====================================================================================
#=====================================================================================

class pioled:
    """ Interact with the luma.oled driver
    """
    def __init__(self):
        global known_fonts
        # Import luma config and instantiate self.device
        if 'create_device args' not in config.sections():
            logging.error (f"Missing section 'create_device args' in the config file - Aborting")
            sys.exit(1)

        self.device_args = SimpleNamespace()
        for key in config.cfg['create_device args']:
            setattr(self.device_args, key, config.cfg['create_device args'][key])

        logging.debug (f"self.device_args: <{self.device_args}>")
        self.device = cmdline.create_device(self.device_args)

        known_fonts = pioled_font_manager()


    def __repr__(self):
        iface = ''
        display_types = cmdline.get_display_types()
        if self.device_args.display not in display_types['emulator']:
            iface = f"Interface: {self.device_args.interface}\n"

        lib_name = cmdline.get_library_for_display_type(self.device_args.display)
        if lib_name is not None:
            lib_version = cmdline.get_library_version(lib_name)
        else:
            lib_name = lib_version = 'unknown'

        import luma.core
        version = f"luma.{lib_name} {lib_version} (luma.core {luma.core.__version__})"
        
        return f"Version: {version}\nDisplay: {self.device_args.display}\n{iface}Dimensions: {self.device.width} x {self.device.height}\n"


#=====================================================================================
#=====================================================================================
#   c l a s s   p i o l e d _ f o n t _ m a n a g e r
#=====================================================================================
#=====================================================================================

class pioled_font_manager:
    """ Create and provide unique combinations of fonts and sizes
    """
    def __init__(self):
        self.known_fonts =  {}
        self.fonts_path =   Path (ir_files(core.tool.main_module)) / "fonts"
        self.default_font = self.fonts_path / config.getcfg('Default_font', DISPLAY_DRIVER_FONT)
        self.default_size = config.getcfg('Default_size', DISPLAY_DRIVER_SIZE)

        if not self.default_font.exists():
            logging.error (f"Default font <{self.default_font} not found - Aborting.")
            sys.exit(1)
        

    def get_font (self, font_name, size):
        font_name_size = font_name + '_' + str(size)
        if not font_name_size in self.known_fonts:
            try:
                self.known_fonts[font_name_size] = ImageFont.truetype(self.fonts_path / font_name, size)
                logging.debug (f"Created known_font <{font_name_size}>")
            except Exception as e:
                logging.warning (f"Failed creating known_font <{font_name_size}> - Using default font and size\n  {type(e).__name__}: {e}")
                self.known_fonts[font_name_size] = ImageFont.truetype(self.fonts_path / self.default_font, self.default_size)
                # TODO any chance of failure?
        
        return self.known_fonts[font_name_size]


#=====================================================================================
#=====================================================================================
#   c l i
#=====================================================================================
#=====================================================================================

def cli():
    """
PiOLED's interactive commands:
- `PiOLED message` displays a multi-line message on the display, supporting list and 
  dictionary formats - see PiOLED.md for format info.  Example:

    PiOLED message "[0, 0, 18, 'Wherever you go...'], {'x':10, 'y':20, 'size':20, 'text':'there you are.', 'font':'ProggyTiny.ttf'}"

- `PiOLED blank`  clears the display
- `PiOLED status` displays the state of the server process, and the ipc semaphores
- `PiOLED unlock` releases the ipc semaphores, which are normally released by the server process
"""
    global config, args
    global logging
    global display_file
    global pioled_logger

    commands = ['message', 'blank', 'status', 'unlock']
    parser = argparse.ArgumentParser(description=cli.__doc__ + __version__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('Command', nargs='?', choices=commands,
                        help=f"Interactive mode Command")
    parser.add_argument('Message', nargs='?',
                        help=f"Message content for Command 'message' - see PiOLED.md for format")

    parser.add_argument('--service', action='store_true',
                        help="Start PiOLED server")
    parser.add_argument('--config-file', '-c', type=str, default=CONFIG_FILE,
                        help=f"Path to the server config file (Default <{CONFIG_FILE}> in user config directory)")
    parser.add_argument('--log-console', '-z', action='store_true',
                        help="Force server logging to the console, overriding config LogFile param")
    parser.add_argument('--val-logfile', default=None,
                        help=f"Path to server log file (abs or rel to core.tool.log_dir_base) used for validation testing (overrides LogFile in config file, default <{None}>)")
    parser.add_argument('--verbose', '-v', action='count', default=0,
                        help="Log status and activity messages (-vv for debug logging)")

    parser.add_argument('--setup-user', action='store_true',
                        help="Install starter files in user space")
    parser.add_argument('--version', '-V', action='version', version=f"{core.tool.toolname} {__version__}",
                        help="Return version number and exit")
    args = parser.parse_args()

    if args.Command is None  and  not args.service  and  not args.setup_user:
        parser.print_help()
        sys.exit()


    # Deploy starter files
    if args.setup_user:
        logging.getLogger('cjnfuncs.deployfiles').setLevel(logging.INFO)
        deploy_files([
            { 'source': CONFIG_FILE,             'target_dir': 'USER_CONFIG_DIR', 'file_stat': 0o644, 'dir_stat': 0o755},
            { 'source': 'PiOLED_server.service', 'target_dir': 'USER_CONFIG_DIR', 'file_stat': 0o644},
            ]) #, overwrite=True)
        sys.exit()


    # Load config file and setup logging
    call_logfile_override = True  if (not args.service  or  args.log_console  or  args.val_logfile)  else False
    try:
        config = config_item(args.config_file)
        config.loadconfig(call_logfile_wins=call_logfile_override, call_logfile=args.val_logfile)
    except Exception as e:
        logging.exception(f"Failed loading config file <{args.config_file}> - Aborting\n  {type(e).__name__}: {e}")
        sys.exit(1)

    display_file = Path(config.getcfg('Display_file'))

    logging.warning (f"========== {core.tool.toolname} ({__version__}), pid {os.getpid()} ==========")
    logging.warning (f"Config file <{config.config_full_path}>, display_file: <{display_file}>")


    # Service/server mode
    if args.service:
        set_logging_level ([logging.WARNING, logging.INFO, logging.DEBUG][args.verbose])    # set server root logger level for service process
        if args.verbose == 2:
            logging.getLogger('cjnfuncs.resourcelock').setLevel(logging.DEBUG)
        if args.val_logfile:
            setuplogging (call_logfile=args.val_logfile, call_logfile_wins=True, FileLogFormat=SERVER_CONS_LOGGING_FORMAT)
        service()                                                                           # service never returns


    # Interactive commands
    if args.Command == 'status':
        import subprocess
        print ("-----------------------------")
        print (f"\n{subprocess.run(['systemctl', 'status', 'PiOLED_server'],  capture_output=True, text=True).stdout}")
        print ("-----------------------------")
        print (f"\nPiOLED_go_flag is currently set?      <{pioled_go_flag.is_locked()}>,  last locked by: <{pioled_go_flag.get_lock_info()}>")
        print (f"\nPiOLED_file_lock is currently locked? <{pioled_file_lock.is_locked()}>,  last locked by: <{pioled_file_lock.get_lock_info()}>")
        print ("-----------------------------")
        sys.exit()

    if args.Command == 'unlock':
        print (f"Before forced unlock:  PiOLED_go_flag is currently set?      <{pioled_go_flag.is_locked()}>")
        pioled_go_flag.unget_lock(force=True, where_called='PiOLED.main - UNLOCK forced unlock')
        print (f"After  forced unlock:  PiOLED_go_flag is currently set?      <{pioled_go_flag.is_locked()}>")
        print (f"Before forced unlock:  PiOLED_file_lock is currently locked? <{pioled_file_lock.is_locked()}>")
        pioled_file_lock.unget_lock(force=True, where_called='PiOLED.main - UNLOCK forced unlock')
        print (f"After  forced unlock:  PiOLED_file_lock is currently locked? <{pioled_file_lock.is_locked()}>")
        sys.exit()

    setuplogging(ConsoleLogFormat=CLI_CONSOLE_LOGGING_FORMAT)
    set_logging_level ([logging.WARNING, logging.INFO, logging.DEBUG][args.verbose], logger_name='cjn_PiTools.PiOLED')
    if args.verbose == 2:
        logging.getLogger('cjnfuncs.resourcelock').setLevel(logging.DEBUG)

    import queue
    pioled_q =  queue.Queue()
    pioled =    pioled_display_driver(pioled_q, display_file=display_file)
    pioled.start()

    if args.Command == 'blank':
        pioled.blank()

    if args.Command == 'message':
        try:
            message = args.Message.strip()
            if not (message.startswith('[')  or  message.startswith('{')):      # 0, 0, 12, 'Wherever you go...'
                message = '[' + message + ']'                                   # [0, 0, 12, 'Wherever you go...']
            message = '[' + message + ']'                                       # [[0, 0, 12, 'Wherever you go...']]
            message = ast.literal_eval(message)                                 # <class 'list'> - [[0, 0, 12, 'Wherever you go...']]
            pioled.message_page(message)
        except Exception as e:
            logging.error (f"Failed to process message <{args.Message}>:\n  {type(e).__name__}: {e}")


if __name__ == '__main__':
    sys.exit(cli())



#!/usr/bin/env python3
"""SHT3x Temperature/RH sensor library for Raspberry Pi
"""
#==========================================================
#
#  Chris Nelson, Copyright 2023-2026
#   
#==========================================================

import time
import logging
from crc import Calculator, Configuration

from .shared import I2C_ERROR, CRC_ERROR, CtoF, FtoC, CtoK, KtoC


# Configs / Constants
TOOLNAME = 'SHT3x'

SHT3x_ADDRS =               [0x44, 0x45]
SOFT_RESET =                [0x30, 0xa2]
FETCH_DATA =                [0xe0, 0x00]
ART =                       [0x2b, 0x32]
READ_STATUS_REGISTER =      [0xf3, 0x2d]
CLEAR_STATUS_REGISTER =     [0x30, 0x41]
HEATER_ENABLE =             [0x30, 0x6d]
HEATER_DISABLE =            [0x30, 0x66]
BREAK =                     [0x30, 0x93]
SINGLE_SHOT_MODES = {
            "High_CS":      [0x2c, 0x06],
            "Medium_CS":    [0x2c, 0x0d],
            "Low_CS":       [0x2c, 0x10],
            "High_noCS":    [0x24, 0x00],
            "Medium_noCS":  [0x24, 0x0b],
            "Low_noCS":     [0x24, 0x16]  }
PERIODIC_DA_MODES = {
            'High_0.5':     [0x20, 0x32],
            'Medium_0.5':   [0x20, 0x24],
            'Low_0.5':      [0x20, 0x2f],
            'High_1':       [0x21, 0x30],
            'Medium_1':     [0x21, 0x26],
            'Low_1':        [0x21, 0x2d],
            'High_2':       [0x22, 0x36],
            'Medium_2':     [0x22, 0x20],
            'Low_2':        [0x22, 0x2b],
            'High_4':       [0x23, 0x34],
            'Medium_4':     [0x23, 0x22],
            'Low_4':        [0x23, 0x29],
            'High_10':      [0x27, 0x37],
            'Medium_10':    [0x27, 0x21],
            'Low_10':       [0x27, 0x2a]  }
READ_ALERT_MODES = {
            'High_Set':     [0xe1, 0x1f],
            'High_Clear':   [0xe1, 0x14],
            'Low_Clear':    [0xe1, 0x09],
            'Low_Set':      [0xe1, 0x02]  }
WRITE_ALERT_MODES = {
            'High_Set':     [0x61, 0x1d],
            'High_Clear':   [0x61, 0x16],
            'Low_Clear':    [0x61, 0x0b],
            'Low_Set':      [0x61, 0x00]  }

MIN_READING_WAIT =  0.001                   # Minimal waiting time in single shot mode between meas trigger and fetch
RESET_WAIT =        0.0015                  # Soft reset spec wait is 1.5ms


crc_calc_config = Configuration(
    width=8,
    polynomial=0x31,
    init_value=0xFF,
    final_xor_value=0x00,
    reverse_input=False,
    reverse_output=False)
crc_calc = Calculator(crc_calc_config)


sht3x_logger = logging.getLogger('cjn_PiTools.SHT3x')


class SHT3x:
    """
## Class SHT3x (device_name, device_addr, pi_i2c_bus_handle, do_reset=True) - SHT3x Temperature/RH sensor library for Raspberry Pi

Create a SHT3x family device instance

### Args
`device_name` (str)
- User defined name for this instance, e.g., 'My_SHT3x'

`device_addr` (int)
- I2C bus address for this instance, e.g., 0x44 or 0x45

`pi_i2c_bus_handle` (cjn_PiTools.shared.pi_i2c instance)
- Get a `pi_i2c` instance handle in the tools script code and pass it to this device instantiation

`do_reset` (bool, default True)
- If True, call `soft_reset()` and `clear_status_reg()` as part of instantiation
- If the device is not accessible at the time of instantiation then set `do_reset=False` and reset the device when available


### Class instance variables - as passed in at instantiation
- `device_name` (str)
- `device_addr` (int)


### Returns
- Handle to the SHT3x instance on success
- Raises ValueError if args checks fail
- Raises RuntimeError if the device fails soft reset or clearing the status register


### Behaviors and rules
- Per the datasheet, "After sending a command to the sensor a minimal waiting time of 1ms is needed before another command
can be received by the sensor." Normally, the time between consecutive API calls takes >1ms, so no padding seems necessary.
- Debug logging may be enabled in the tool script code by setting this module's logging level:

        logging.getLogger('cjn_PiTools.SHT3x').setLevel(logging.DEBUG)
"""

    def __init__(self, device_name, device_addr, pi_i2c_bus_handle, do_reset=True):
        self.device_name =          device_name
        self.device_addr =          device_addr
        self.pi_i2c_bus_handle =    pi_i2c_bus_handle

        if not isinstance(self.device_name, str):
            raise ValueError (f"SHT3x device_name must be a str.  Received <{device_name}>")

        if self.device_addr not in SHT3x_ADDRS:
            raise ValueError (f"<{self.device_name}> SHT3x device address must be 0x44 or 0x45.  Received <0x{device_addr:0>2x}>")

        if do_reset:
            if self.soft_reset() == I2C_ERROR:
                raise RuntimeError (f"<{self.device_name}> soft_reset() during instantiation failed")

            if self.clear_status_reg() == I2C_ERROR:
                raise RuntimeError (f"<{self.device_name}> clear_status_reg() during instantiation failed")

        api = 'smbus'  if self.pi_i2c_bus_handle.api == 'smbus'  else 'pigpio'
        sht3x_logger.debug (f"<{self.device_name}> New SHT3x device defined at addr <0x{self.device_addr:0>2x}> using api <{api}> on i2c bus <{self.pi_i2c_bus_handle.i2c_bus_num}>")


    #=====================================================================================
    #=====================================================================================
    #  s o f t _ r e s e t
    #=====================================================================================
    #=====================================================================================

    def soft_reset(self, reset_wait=RESET_WAIT):
        """
## soft_reset (reset_wait=0.0015) - Issue a soft_reset

***SHT3x class member function***


### Args
`reset_wait` (float, default 1.5ms)
- The spec time is 1.5ms


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- The reset time is blocking to the calling code
- A soft_reset() does NOT clear the status register
- A soft_reset() DOES reset the alert High/Low Set/Clear registers to their defaults
- From the datasheet:  "This triggers the sensor to reset its system controller and reloads calibration data from the memory.
It is worth noting that the sensor reloads calibration data prior to every measurement by default."  Note that a soft_reset 
cannot be applied if the system controller (I2C interface) is stuck - a power cycle is required if your device version does not
have a reset pin.
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** soft_reset()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, SOFT_RESET)
            time.sleep (reset_wait)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  s i n g l e _ s h o t
    #=====================================================================================
    #=====================================================================================

    def single_shot(self, tempunits='C', repeatability='High', reading_wait=-1, fetch_data=True):
        """
## single_shot (tempunits='C', repeatability='High', reading_wait=-1, fetch_data=True) - Issue a single shot temperature/RH measurement and read back the results

***SHT3x class member function***


### Args
`tempunits` (str, default 'C', case-independent)
- Must be 'C', 'F' or 'K'

`repeatability` (str, default 'High')
- Select repeatability setting 'High', 'Medium', or 'Low' per datasheet, with corresponding measurement time differences

`reading_wait` (int or float, default -1)
- If = -1 then use clock stretch mode (device indicates when data is available by releasing TCK)
- If >= 0 then use non-clock-stretch mode and use `reading_wait` value between triggering the single shot 
measurement and fetching the data (if `fetch_data` = True)
- The worst case maximum measurement durations are High repeatability 15.5ms, Medium 6.5ms, and Low 4.5ms. 
See the datasheet.

`fetch_data` (bool, default True)
- If True, the temperature and RH data is read back after the measurement is triggered
- If False, the measurement is triggered but the data is not read back.  A separate call to `fetch_data()` may be used to 
read back the measurement results.


### Returns
- If `fetch_data` = True, return (Temperature, RH) tuple on success
- If `fetch_data` = False, return (0, 0) tuple on success
- (I2C_ERROR, I2C_ERROR) tuple on I2C IO error
- (CRC_ERROR, CRC_ERROR) tuple on fetched data CRC mismatch
- Raises ValueError on invalid args


### Behaviors and rules
- The single shot measurement trigger and fetching the temperature/rh results may be separated by setting `reading_wait = 0`
(no delay after the trigger) and `fetch_data = False`.  Follow this single shot trigger call with a later call to 
`fetch_data()` to get the results.  Consider the minimum measurement time rules before calling `fetch_data()`.
- The entire measurement sequence (trigger single shot, any `reading_wait` or clock stretch time, and the `fetch_data()` call) 
is blocking to the calling code.
- If the devices is still busy doing the measurement when `fetch_data()` is called (`reading_wait` too short) then 
(I2C_ERROR, I2C_ERROR) is returned and debug logging will show:
  - smbus api:  OSError: [Errno 121] Remote I/O error
  - pigpio api: OSError: i2c_read_device failed - error code <-83>
"""

        if not isinstance(reading_wait, (int, float))  or  not (reading_wait == -1  or  reading_wait >= 0.0):
            raise ValueError (f"<{self.device_name}> Invalid reading_wait value - expecting -1 or >= 0.0 - received <{reading_wait}>")

        mode = '_CS'  if reading_wait == -1  else '_noCS'
        try:
            bytes_code = SINGLE_SHOT_MODES[repeatability + mode]
        except:
            raise ValueError (f"<{self.device_name}> Invalid Single Shot mode selection - received repeatability <{repeatability}>")

        sht3x_logger.debug (f"<{self.device_name}> ***** single_shot()  <{repeatability}{mode}>")

        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, bytes_code)
            if reading_wait == -1:
                time.sleep(MIN_READING_WAIT)
            else:
                time.sleep(reading_wait)
            if fetch_data:
                return self.fetch_data(tempunits, send_fetch=False)
            else:
                return 0, 0
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR, I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  s t a r t _ p e r i o d i c _ D A
    #=====================================================================================
    #=====================================================================================

    def start_periodic_DA (self, repeatability='High', mps=2):
        """
## start_periodic_DA (repeatability='High', mps=2) - Start Periodic Data Acquisition Mode

***SHT3x class member function***


### Args
`repeatability` (str, default 'High')
- Select repeatability setting 'High', 'Medium', or 'Low' per datasheet, with corresponding measurement time differences

`mps` (int or float, default 2)
- Measurements per second.
- Allowed values are 0.5, 1, 2, 4, and 10


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if `repeatability` or `mps` is invalid


### Behaviors and rules
- Status bit 0x0020 indicates Periodic / free running measurement mode is active (undocumented)
- Status bit 0x0040 indicates new measurement data is available (undocumented)
- Attempting to fetch data more frequently than new data is available (based on the `mps` setting) results in 
`fetch_data()` returning (I2C_ERROR, I2C_ERROR) and debug logging will show:
  - smbus api:  OSError: [Errno 121] Remote I/O error
  - pigpio api: OSError: i2c_read_device failed - error code <-83>
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** start_periodic_DA()")
        try:
            bytes_code = PERIODIC_DA_MODES[repeatability + '_' + str(mps)]
        except:
            raise ValueError (f"<{self.device_name}> Invalid Periodic Data Acquisition mode selection - received repeatability <{repeatability}>, mps: <{mps}>")

        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, bytes_code)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  s t a r t _ A R T
    #=====================================================================================
    #=====================================================================================

    def start_ART (self):
        """
## start_ART () - Start Periodic Data Acquisition Mode using Accelerated Response Time

***SHT3x class member function***


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- New measurement data is available at 4 measurements per second (250ms)
- Status bit 0x0020 indicates Periodic / free running measurement mode is active (undocumented)
- Status bit 0x0040 indicates new measurement data is available (undocumented)
- Attempting to fetch data more frequently than new data is available results in 
`fetch_data()` returning (I2C_ERROR, I2C_ERROR) and debug logging will show:
  - smbus api:  OSError: [Errno 121] Remote I/O error
  - pigpio api: OSError: i2c_read_device failed - error code <-83>
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** start_ART()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, ART)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  s t o p _ p e r i o d i c _ D A
    #=====================================================================================
    #=====================================================================================

    def stop_periodic_DA (self):
        """
## stop_periodic_DA () - Stop Periodic Data Acquisition Mode, including ART mode

***SHT3x class member function***


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** stop_periodic_DA()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, BREAK)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  f e t c h _ d a t a
    #=====================================================================================
    #=====================================================================================

    def fetch_data(self, tempunits='C', send_fetch=True, force_CRC_fail=False):
        """
## fetch_data (tempunits='C', send_fetch=True, force_CRC_fail=False) - Fetch temperature and RH data from the device

***SHT3x class member function***


### Args
`tempunits` (str, default 'C', case-independent)
- Must be 'C', 'F' or 'K'

`send_fetch` (bool, default True)
- If True, the fetch data instruction will be sent to the device
- `single_shot()` calls `fetch_data()` with `send_fetch=False` since the device is ready to be read 
after the single shot measurement trigger has completed.
- When using a Periodic Data Acquisition Mode call `fetch_data()` with `send_fetch=True` (default)

`force_CRC_fail` (bool, default False)
- Used for validation


### Returns
- (Temperature, RH) tuple on success
- (I2C_ERROR, I2C_ERROR) tuple on I2C IO error
- (CRC_ERROR, CRC_ERROR) tuple on fetched data CRC mismatch
- Raises ValueError if `tempunits` is invalid


### Behaviors and rules
- Status bit 0x0040 indicates new measurement data is available (undocumented)
- Attempting to fetch data when no new data is available results in 
`fetch_data()` returning (I2C_ERROR, I2C_ERROR) and debug logging will show:
  - smbus api:  OSError: [Errno 121] Remote I/O error
  - pigpio api: OSError: i2c_read_device failed - error code <-83>
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** fetch_data()")
        try:
            if send_fetch:
                self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, FETCH_DATA)

            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 6)

        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR, I2C_ERROR
        
        if count != 6:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR - error code <{count}>")
            return I2C_ERROR, I2C_ERROR

        if force_CRC_fail:
            data[2] = data[5] = 0x00

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            xx =  f"<{self.device_name}> Temp/RH raw data:\n"
            xx += f"  temp data returned bytes: 0x{data[0]:0>2x} 0x{data[1]:0>2x} 0x{data[2]:0>2x},  calc CRC: <0x{crc_calc.checksum(bytes([data[0], data[1]])):0>2x}>\n"
            xx += f"  RH   data returned bytes: 0x{data[3]:0>2x} 0x{data[4]:0>2x} 0x{data[5]:0>2x},  calc CRC: <0x{crc_calc.checksum(bytes([data[3], data[4]])):0>2x}>"
            sht3x_logger.debug (xx)

        # Process temp data
        if crc_calc.checksum(bytes([data[0], data[1]])) != data[2]:
            sht3x_logger.debug (f"<{self.device_name}> temp data CRC error")
            temp_rslt = CRC_ERROR
        else:
            rawtemp = ((data[0] <<8) | data[1]) & 0xFFFF
            temp_rslt = _decode_temp (rawtemp)
            _tempunits = tempunits.lower()
            if _tempunits not in ['c', 'f', 'k']:
                raise ValueError (F"<{self.device_name}> tempunits must be 'C', 'F', or 'K' - received <{tempunits}>")

            if _tempunits == 'f':
                temp_rslt = CtoF(temp_rslt)
            elif _tempunits == 'k':
                temp_rslt = CtoK(temp_rslt)

        sht3x_logger.debug (f"<{self.device_name}> Calculated temp ({tempunits}):     {temp_rslt:>5.1f}")

        # Process RH data
        if crc_calc.checksum(bytes([data[3], data[4]])) != data[5]:
            sht3x_logger.debug (f"<{self.device_name}> RH data CRC error")
            RH_rslt = CRC_ERROR
        else:
            rawRH = ((data[3] <<8) | data[4]) & 0xFFFF
            RH_rslt = _decode_rh(rawRH)

        sht3x_logger.debug (f"<{self.device_name}> Calculated RH:           {RH_rslt:>5.1f}")

        return temp_rslt, RH_rslt


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ s t a t u s _ r e g
    #=====================================================================================
    #=====================================================================================

    def read_status_reg(self, quiet=False, force_CRC_fail=False):
        """
## read_status_reg (quiet=False, force_CRC_fail=False) - Read and return the status register

***SHT3x class member function***


### Args
`quiet` (bool, default False)
- Set True by SHT3x class internal calls for debug logging

`force_CRC_fail` (bool, default False)
- Used for validation


### Returns
- 2-byte status register value (16-bit int) on success
- I2C_ERROR on I2C IO error
- CRC_ERROR on fetched data CRC mismatch


### Behaviors and rules
- If `quiet=False` and module debug logging is enabled the status register bits are decoded and logged
- Status register bit 4 (reset detected) does not seem to be set by a soft_reset(), contrary to the datasheet.
"""

        if not quiet:
            sht3x_logger.debug (f"<{self.device_name}> ***** read_status_reg()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, READ_STATUS_REGISTER)
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 3)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if count != 3:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR - error code <{count}>")
            return I2C_ERROR

        if force_CRC_fail:
            data[2] = 0x00

        if not quiet:
            sht3x_logger.debug (f"<{self.device_name}> status reg raw data: 0x{data[0]:0>2x} 0x{data[1]:0>2x} 0x{data[2]:0>2x},  calc CRC: <0x{crc_calc.checksum(bytes([data[0], data[1]])):0>2x}>")

        if crc_calc.checksum(bytes([data[0], data[1]])) != data[2]:
            sht3x_logger.debug (f"<{self.device_name}> status register data CRC error")
            return CRC_ERROR
        
        reg_value = ((data[0] <<8) | data[1]) & 0xFFFF
        if not quiet:
            if reg_value & 0x8000:
                sht3x_logger.info ("  At least one pending alert")
            if reg_value & 0x4000:
                sht3x_logger.info ("  Reserved (undocumented)")
            if reg_value & 0x2000:
                sht3x_logger.info ("  Heater ON")
            if reg_value & 0x1000:
                sht3x_logger.info ("  Reserved (undocumented)")
            if reg_value & 0x0800:
                sht3x_logger.info ("  RH tracking alert")
            if reg_value & 0x0400:
                sht3x_logger.info ("  T  tracking alert")
            if reg_value & 0x0200:
                sht3x_logger.info ("  Reserved (undocumented)")
            if reg_value & 0x0100:
                sht3x_logger.info ("  Reserved (undocumented)")
            if reg_value & 0x0080:
                sht3x_logger.info ("  New data available - stale data overwritten? (undocumented)")
            if reg_value & 0x0040:
                sht3x_logger.info ("  New data available? (undocumented)")
            if reg_value & 0x0020:
                sht3x_logger.info ("  PeriodicDA running? (undocumented)")
            if reg_value & 0x0010:
                sht3x_logger.info ("  System reset detected")
            if reg_value & 0x0008:
                sht3x_logger.info ("  Reserved (undocumented)")
            if reg_value & 0x0004:
                sht3x_logger.info ("  Reserved (undocumented)")
            if reg_value & 0x0002:
                sht3x_logger.info ("  Last command not processed")
            if reg_value & 0x0001:
                sht3x_logger.info ("  Checksum of last write transfer failed")

        return reg_value


    #=====================================================================================
    #=====================================================================================
    #  c l e a r _ s t a t u s _ r e g
    #=====================================================================================
    #=====================================================================================

    def clear_status_reg(self):
        """
## clear_status_reg () - Clear the status register

***SHT3x class member function***


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- `clear_status_reg()` clears only the alert status bits 15, 11, and 10, and bit 4 (System reset detected).
Other bits, including undocumented bits, are not cleared.
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** clear_status_reg()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, CLEAR_STATUS_REGISTER)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  h e a t e r _ e n a b l e
    #=====================================================================================
    #=====================================================================================

    def heater_enable(self):
        """
## heater_enable () - Turn the heater on

***SHT3x class member function***


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** heater_enable()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, HEATER_ENABLE)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  h e a t e r _ d i s a b l e
    #=====================================================================================
    #=====================================================================================

    def heater_disable(self):
        """
## heater_disable () - Turn the heater off

***SHT3x class member function***


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
"""

        sht3x_logger.debug (f"<{self.device_name}> ***** heater_disable()")
        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, HEATER_DISABLE)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        if sht3x_logger.isEnabledFor(logging.DEBUG):
            sht3x_logger.debug (f"<{self.device_name}> Status reg:  0x{self.read_status_reg(quiet=True):0>4x}")
        return 0


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ a l e r t _ r e g
    #=====================================================================================
    #=====================================================================================

    def read_alert_reg(self, reg_select, tempunits='C', force_CRC_fail=False):
        """
## read_alert_reg (reg_select, tempunits='C', force_CRC_fail=False) - Fetch temperature and RH alert values from selected register

***SHT3x class member function***


### Args
`reg_select` (str)
- One of 'High_Set', 'High_Clear', 'Low_Clear', or 'Low_Set'

`tempunits` (str, default 'C', case-independent)
- Must be 'C', 'F' or 'K'

`force_CRC_fail` (bool, default False)
- Used for validation


### Returns
- (Temperature, RH) tuple on success
- (I2C_ERROR, I2C_ERROR) tuple on I2C IO error
- (CRC_ERROR, CRC_ERROR) tuple on fetched data CRC mismatch
- Raises ValueError if `reg_select` or `tempunits` is invalid
"""

        try:
            reg_code = READ_ALERT_MODES[reg_select]
        except:
            raise ValueError (f"<{self.device_name}> Invalid Alert Register selection - received <{reg_select}>")
        sht3x_logger.debug (f"<{self.device_name}> ***** read_alert_reg() <{reg_select}>")

        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, reg_code)
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 3)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR, I2C_ERROR

        if count != 3:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR - error code <{count}>")
            return I2C_ERROR, I2C_ERROR

        if force_CRC_fail:
            data[2] = 0x00
        if crc_calc.checksum(bytes([data[0], data[1]])) != data[2]:
            sht3x_logger.debug (f"<{self.device_name}> Alert register <{reg_select}> data CRC error")
            return CRC_ERROR, CRC_ERROR
        
        temp, rh = self._decode_alert_word (data[0] <<8 | data[1], tempunits)
        sht3x_logger.debug (f"<{self.device_name}> <{reg_select}> temp: <{temp:5.1f}{tempunits}>, rh: <{rh:5.1f}%>")

        return temp, rh


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ a l e r t _ r e g
    #=====================================================================================
    #=====================================================================================

    def write_alert_reg(self, reg_select, temp, rh, tempunits='C', force_CRC_fail=False):
        """
## write_alert_reg (reg_select, temp, rh, tempunits='C', force_CRC_fail=False) - Write temperature and RH alert values to selected register

***SHT3x class member function***


### Args
`reg_select` (str)
- One of 'High_Set', 'High_Clear', 'Low_Clear', or 'Low_Set'

`temp` (int or float)
- Temperature value in `tempunits`
- Must be in range of -40C to 125C

`rh` (int or float)
- Relative humidity value in %
- Must be in range of 0% to 100%

`tempunits` (str, default 'C', case-independent)
- Must be 'C', 'F' or 'K'

`force_CRC_fail` (bool, default False)
- Used for validation


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if `reg_select` or `tempunits` are invalid, or `temp` or `rh` are out of range


### Behaviors and rules
- A CRC error on the write data to the alert registers causes a stuck I2C bus, requiring a power cycle to recover.
"""

        try:
            reg_code = WRITE_ALERT_MODES[reg_select]
        except:
            raise ValueError (f"<{self.device_name}> Invalid Alert Register selection - received <{reg_select}>")

        if tempunits.lower() == 'c':
            tempC = temp
        elif tempunits.lower() == 'f':
            tempC = FtoC(tempC)
        elif tempunits.lower() == 'k':
            tempC = KtoC(tempC)
        else:
            raise ValueError (F"<{self.device_name}> tempunits must be 'C', 'F', or 'K' - received <{tempunits}>")

        if tempC < -40.0  or  tempC > 125.0:
            raise ValueError (f"<{self.device_name}> Invalid temperature value - Expecting -40C <= temp value <= 125C - received <{tempC}C>")

        if rh < 0.0  or  rh > 100.0:
            raise ValueError (f"Invalid RH value - Expecting 0.0 <= rh value <= 100.0 - received <{rh}>")

        tempbits =  _encode_temp(tempC) >> 7
        rhbits =    _encode_rh(rh) & 0xfe00
        regbits = rhbits | tempbits
        sht3x_logger.debug (f"<{self.device_name}> ***** write_alert_reg() <{reg_select}>: <{self._decode_alert_word(regbits, tempunits)}>")
        MSB = regbits >> 8
        LSB = regbits & 0xff
        regCRC = crc_calc.checksum(bytes([MSB, LSB]))  if not force_CRC_fail  else 0x00

        payload = reg_code
        payload.append(MSB)
        payload.append(LSB)
        payload.append(regCRC)

        try:
            self.pi_i2c_bus_handle.i2c_write_device(self.device_addr, payload)
        except Exception as e:
            sht3x_logger.debug (f"<{self.device_name}> I2C_ERROR\n  {type(e).__name__}: {e}")
            return I2C_ERROR

        return 0

    #---------------------------------------------------------------------
    def _decode_alert_word (self, word, tempunits):
        # Upper 7 bits are the RH value MSBs
        # Lower 9 bits are the Temp value MSBs
        # Returns tuple (temp (in C or F), rh)

        tempbits = (word & 0x01ff) << 7
        temp = _decode_temp(tempbits)

        _tempunits = tempunits.lower()
        if _tempunits not in ['c', 'f', 'k']:
            raise ValueError (F"<{self.device_name}> tempunits must be 'C', 'F', or 'K' - received <{tempunits}>")

        if _tempunits == 'f':
            temp = CtoF(temp)
        elif _tempunits == 'k':
            temp = CtoK(temp)

        rh = _decode_rh (word & 0xfe00)
        return temp, rh


def _decode_temp (tempbits):     # returns tempC
    return -45.0 + (175 * tempbits / 65535)   #(2**16 -1))

def _encode_temp (tempC):
    return int((tempC + 45.0) / (175/65535))

def _decode_rh (rhbits):
    return 100.0 * rhbits / 65535

def _encode_rh (rh):
    return int(rh / 100.0 * 65535)



#!/usr/bin/env python3
"""Shared functions library for cjn_PiTools
"""

#==========================================================
#
#  Chris Nelson, Copyright 2026
#
#==========================================================

import importlib.metadata
__version__ = importlib.metadata.version(__package__ or __name__)

import math
import logging

# Configs / Constants
BUS0_I2C_SDA_GPIO = 0
BUS0_I2C_SCL_GPIO = 1
BUS1_I2C_SDA_GPIO = 2
BUS1_I2C_SCL_GPIO = 3

I2C_ERROR =         -256
CRC_ERROR =         -255

logging.getLogger('cjn_PiTools').setLevel(logging.WARNING)  # Set default log level for all cjn_PiTools modules
shared_logger = logging.getLogger('cjn_PiTools.shared')


#=====================================================================================
#=====================================================================================
#  C l a s s   p i _ i 2 c
#=====================================================================================
#=====================================================================================

class pi_i2c:
    """
## Class pi_i2c (api, i2c_bus_num=1) - Get a pi_i2c handle for the specified I2C bus and api

Supports both smbus and pigpio I2C APIs


### Args
`api` (pigpio handle or str 'smbus')
- A pigpio handle is returned from `pigpio.pi()`
- If not 'smbus' then `api` is assumed to be a valid pigpio handle (not validity checked)

`i2c_bus_num` (int 0 or 1, default 1)
- Bus 1 is the typical user I2C bus
- Bus 0 is often reserved for comm with HAT boards, but may be used as well
- `i2c_bus_num` is not validity checked


### Returns
- pi_i2c handle for I2C communication using the specified api and bus
    """

    def __init__(self, api, i2c_bus_num=1):
        global i2c_msg
        self.api =                  api
        self.i2c_bus_num =          i2c_bus_num
        
        sda = BUS0_I2C_SDA_GPIO  if self.i2c_bus_num == 0  else BUS1_I2C_SDA_GPIO
        scl = BUS0_I2C_SCL_GPIO  if self.i2c_bus_num == 0  else BUS1_I2C_SCL_GPIO
        self.i2c_device_handles =   {}

        if self.api == 'smbus':
            shared_logger.debug (f"Set up <smbus> mode for i2c_bus_num <{self.i2c_bus_num}>")
            from smbus2.smbus2 import SMBus, i2c_msg
            self.smbus_handle = SMBus(self.i2c_bus_num)
            pass
        else:   # pigpio API
            shared_logger.debug (f"Set up <pigpio> mode for i2c_bus_num <{self.i2c_bus_num}>")
            import pigpio
            self.api.set_mode(sda, pigpio.ALT0)                 # set pins to I2C mode
            self.api.set_mode(scl, pigpio.ALT0)


    #=====================================================================================
    #=====================================================================================
    #  i 2 c _ w r i t e _ d e v i c e
    #=====================================================================================
    #=====================================================================================

    def i2c_write_device (self, addr, bytes_list):
        """
## i2c_write_device (addr, bytes_list) - Write a list of bytes to the target device

***pi_i2c class member function***


### Args
`addr` (int)
- i2c address of target device - range 0x00 to 0x7F
- No validity checks - the caller should confirm address validity

`bytes_list` (list of bytes)
- Data to be written to the target device, e.g., `[0x05, 0xFF]`
- Minimal data validity checking - some invalid data results in write failure


### Returns
- On success, returns number of bytes written (length of `bytes_list`)
- Raises ValueError if `bytes_list` is not a list or is an empty list.
- On fail, raises exception (actually raised by underlying `api` call, e.g., I2C IO error, or invalid data in `bytes_list`)


### Behaviors and rules
- Commonly, the first byte in the `bytes_list` is taken as a register address to which the following bytes are written. 
  Note that this behavior is device-type specific.  The contents of `bytes_list` will simply be sent out after 
  the device at `addr` has been addressed for writing.
- Uses pigpio i2c_write_device() and smbus write_i2c_block_data()
        """

        if not isinstance(bytes_list, list)  or  len(bytes_list) == 0:
            raise ValueError (f"Illegal bytes_list received: <{bytes_list}>")
        if shared_logger.isEnabledFor(logging.DEBUG):
            shared_logger.debug (f"Using api <{self.api}> write bus <{self.i2c_bus_num}> addr <0x{addr:0>2x}> data <{_bytes_to_str(bytes_list)}>")
        if self.api == 'smbus':
            b0 = bytes_list[0]
            b1plus = bytes_list[1:]
            self.smbus_handle.write_i2c_block_data(addr, b0, b1plus)    # returns None
            return len(bytes_list)
        else:   # pigpio API
            i2c_device_handle = self._get_pigpio_i2c_device_handle(addr)
            xx = self.api.i2c_write_device(i2c_device_handle, bytes_list)
            if xx < 0:
                raise OSError (f"i2c_write_device failed - error code <{xx}>")
            return len(bytes_list)


    #=====================================================================================
    #=====================================================================================
    #  i 2 c _ r e a d _ d e v i c e
    #=====================================================================================
    #=====================================================================================

    def i2c_read_device (self, addr, read_byte_count):
        """
## i2c_read_device (addr, read_byte_count) - Read a number of bytes from the target device

***pi_i2c class member function***


### Args
`addr` (int)
- i2c address of target device - range 0x00 to 0x7F
- No validity checks - caller should confirm address validity

`read_byte_count` (int)
- Number of bytes to read from the target device
- No validity checking


### Returns
- On success, returns tuple (actual read count, [data]), e.g., `(3, [0x01, 0x02, 0x03])`
- On fail, raises exception (actually raised by underlying `api` call, e.g., I2C IO error, OSError, etc.).
- If using the pigpio api, if the returned byte count is negative this error code is passed within a raised OSError exception.


### Behaviors and rules
- Uses pigpio i2c_read_device() and smbus i2c_msg.read() / i2c_rdwr
        """

        if shared_logger.isEnabledFor(logging.DEBUG):
            shared_logger.debug (f"Using api <{self.api}> read bus <{self.i2c_bus_num}> addr <0x{addr:0>2x}> read_byte_count <{read_byte_count}>")
        if self.api == 'smbus':
            msg = i2c_msg.read(addr, read_byte_count)
            self.smbus_handle.i2c_rdwr(msg)
            return read_byte_count, list(msg)
        else:   # pigpio API
            i2c_device_handle = self._get_pigpio_i2c_device_handle(addr)
            count, data = self.api.i2c_read_device(i2c_device_handle, read_byte_count)
            if count < 0:
                raise OSError (f"i2c_read_device failed - error code <{count}>")
            return count, list(data)


    #=====================================================================================
    #=====================================================================================
    #  i 2 c _ w r i t e _ b y t e
    #=====================================================================================
    #=====================================================================================

    def i2c_write_byte (self, addr, byte_value):
        """
## i2c_write_byte (addr, byte_value) - Write one byte to the target device

***pi_i2c class member function***


### Args
`addr` (int)
- i2c address of target device - range 0x00 to 0x7F
- No validity checks - caller should confirm address validity

`byte_value` (int)
- Data to be written to the target device, e.g., `0x55`
- The value is check to be an int in the range of 0x00 to 0xFF


### Returns
- On success, returns 1 (meaning one byte written)
- Raises ValueError if `byte_value` is not an int in the range of 0x00 to 0xFF
- On fail, raises exception (actually raised by underlying `api` call, e.g., I2C IO error, OSError, etc.).


### Behaviors and rules
- Uses pigpio i2c_write_byte() and smbus write_byte()
        """

        if not isinstance(byte_value, int)  or  byte_value < 0x00  or  byte_value > 0xFF:
            raise ValueError (f"Illegal byte_value received: <{byte_value}>")
        if shared_logger.isEnabledFor(logging.DEBUG):
            shared_logger.debug (f"Using api <{self.api}> write bus <{self.i2c_bus_num}> addr <0x{addr:0>2x}> data <{_bytes_to_str(byte_value)}>")

        if self.api == 'smbus':
            rslt = self.smbus_handle.write_byte(addr, byte_value)
        else:   # pigpio API
            i2c_device_handle = self._get_pigpio_i2c_device_handle(addr)
            rslt = self.api.i2c_write_byte(i2c_device_handle, byte_value)
            if rslt < 0:
                raise OSError (f"i2c_write_device failed - error code <{rslt}>")
        return 1


    #=====================================================================================
    #=====================================================================================
    #  i 2 c _ r e a d _ b y t e
    #=====================================================================================
    #=====================================================================================

    def i2c_read_byte (self, addr):
        """
## i2c_read_byte (addr) - Read one byte from the target device

***pi_i2c class member function***


### Args
`addr` (int)
- i2c address of target device - range 0x00 to 0x7F
- No validity checks - caller should confirm address validity


### Returns
- On success, returns one byte read from the target device, e.g., `0xAA`
- On fail, raises exception (actually raised by underlying `api` call, e.g., I2C IO error, OSError, etc.).


### Behaviors and rules
- Uses pigpio i2c_read_byte() and smbus read_byte()
        """

        if shared_logger.isEnabledFor(logging.DEBUG):
            shared_logger.debug (f"Using api <{self.api}> read 1 byte from bus <{self.i2c_bus_num}> addr <0x{addr:0>2x}>")

        if self.api == 'smbus':
            byte_value = self.smbus_handle.read_byte(addr)
        else:   # pigpio API
            i2c_device_handle = self._get_pigpio_i2c_device_handle(addr)
            byte_value = self.api.i2c_read_byte(i2c_device_handle)

        if shared_logger.isEnabledFor(logging.DEBUG):
            shared_logger.debug (f"i2c_read_byte from addr <0x{addr:0>2x}>:  <{_bytes_to_str(byte_value)}>")
        return byte_value


    #=====================================================================================
    #=====================================================================================
    #  c l o s e
    #=====================================================================================
    #=====================================================================================

    def close(self):
        """
## close () - Close all connections associated with this pi_i2c handle

***pi_i2c class member function***


### Returns
- Returns 0


### Behaviors and rules
- If using pigpio, the pigpio handle (`pio = pigpio.pi()` in the above example) is owned by and must be stopped in the tool script code.
The pi_i2c.close() method only closes the allocated i2c device handles within the pigpio connection.
        """

        if self.api == 'smbus':
            shared_logger.debug (f"smbus_handle closed")
            self.smbus_handle.close()
        else:   # pigpio API
            for addr in self.i2c_device_handles:
                i2c_device_handle = self.i2c_device_handles[addr]
                shared_logger.debug (f"Closing i2c_device_handle <{i2c_device_handle}>")
                self.api.i2c_close(i2c_device_handle)
        return 0


    #=====================================================================================
    #=====================================================================================
    #  _ g e t _ p i g p i o _ i 2 c _ d e v i c e _ h a n d l e
    #=====================================================================================
    #=====================================================================================

    # Private method

    def _get_pigpio_i2c_device_handle (self, addr):
        """_get_pigpio_i2c_device_handle manages the allocation and reuse of pigpio i2c device handles
        """
        if not addr in self.i2c_device_handles:
            i2c_device_handle = self.i2c_device_handles[addr] = self.api.i2c_open(self.i2c_bus_num, addr)
            shared_logger.debug (f"New i2c_device_handle: <{i2c_device_handle}> for addr <0x{addr:0>2x}>")
        else:
            i2c_device_handle = self.i2c_device_handles[addr]
            shared_logger.debug (f"Got i2c_device_handle: <{i2c_device_handle}> for addr <0x{addr:0>2x}>")
        return i2c_device_handle


#=====================================================================================
#=====================================================================================
#  T e m p e r a t u r e   c o n v e r s i o n
#=====================================================================================
#=====================================================================================

def CtoF(tempC):
    """
## CtoF (tempC) - Convert temperature value in Celsius to Fahrenheit

### Arg
`tempC` (float)
- Temperature value in Celsius
- Value not validated

### Returns
- Returns float temperature value in Fahrenheit
    """
    return tempC*1.8 +32.0


def FtoC(tempF):
    """
## FtoC (tempF) - Convert temperature value in Fahrenheit to Celsius

### Arg
`tempF` (float)
- Temperature value in Fahrenheit
- Value not validated

### Returns
- Returns float temperature value in Celsius
    """
    return (tempF -32.0) / 1.8


def CtoK(tempC):
    """
## CtoK (tempC) - Convert temperature value in Celsius to Kelvin

### Arg
`tempC` (float)
- Temperature value in Celsius
- Value not validated

### Returns
- Returns float temperature value in Kelvin
    """
    return tempC + 273.15


def KtoC(tempK):
    """
## KtoC (tempK) - Convert temperature value in Kelvin to Celsius

### Arg
`tempK` (float)
- Temperature value in Kelvin
- Value not validated

### Returns
- Returns float temperature value in Celsius
    """
    return tempK - 273.15


def calculate_dew_point(tempC, RH):
    """
## calculate_dew_point (tempC, RH) - Calculate the dew point given tempC and relative humidity

Uses the Mangus formula - [Wikipedia](https://en.wikipedia.org/wiki/Dew_point)

### Arg
`tempC` (float)
- Temperature value in Celsius
- Value not validated

RH (float)
- Relative humidity in percent
- Value not validated

### Returns
- Returns dew point temperature value in Celsius.  Pass the returned temperature thru CtoF() if the dew
point in Fahrenheit is needed, e.g., `dp_f = CtoF(calculate_dew_point(22.6, 31.2))`
    """
    # Magnus formula constants
    a = 17.62
    b = 243.12

    # Calculate gamma
    gamma = (a * tempC) / (b + tempC) + math.log(RH / 100.0)

    # Calculate dew point in Celsius
    dew_point_c = (b * gamma) / (a - gamma)
    return dew_point_c


#=====================================================================================
#=====================================================================================
#  P r i v a t e   f u n c t i o n s
#=====================================================================================
#=====================================================================================

def _bytes_to_str(_bytes):
    # Expects one byte or list of bytes
    # Returns str hex representation
    # If can't convert, then return _bytes

    try:
        if isinstance(_bytes, int):
            return f"0x{_bytes:0>2x}"
        else:   # Must be list of bytes
            xx = '['
            for item in _bytes:
                xx += f" 0x{item:0>2x}"
            return xx + ' ]'
    except:
        return _bytes
    

# Remnant
    #=====================================================================================
    #=====================================================================================
    #  i2c_read_i2c_block_data
    #=====================================================================================
    #=====================================================================================

    # def i2c_read_i2c_block_data (self, addr, reg, read_byte_count):
    #     if self.api == 'smbus':
    #         msg = i2c_msg.read(addr, read_byte_count)
    #         self.smbus_handle.i2c_rdwr(msg)
    #         return read_byte_count, list(msg)
    #     else:   # pigpio API
    #         i2c_device_handle = self.get_pigpio_i2c_device_handle(addr)
    #         try:
    #             count, data = self.api.i2c_read_i2c_block_data(i2c_device_handle, reg, read_byte_count)
    #             # if count < 0:
    #             #     raise OSError (f"i2c_read_device failed - error code <{count}>")
    #             # return count, data
    #         except Exception as e:
    #             raise OSError (f"i2c_read_i2c_block_data failed (pigpio exception: {type(e).__name__}: {e})")
    #         if count < 0:
    #             raise OSError (f"i2c_read_i2c_block_data failed - error code <{count}>")
    #         return count, data





#!/usr/bin/env python3
"""PiBlinky - A threaded multiple LED driver for Raspberry Pi
"""

import importlib.metadata
__version__ = importlib.metadata.version(__package__ or __name__)

#==========================================================
#
#  Chris Nelson, Copyright 2022-2026
#
#==========================================================

import time
import sys
import logging
from threading import Thread

# Configs / Constants
CMD_SAVE    = 1
CMD_RESTORE = 2
CMD_EXIT    = -99
GPIO_NUMBERS = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27]

piblinky_logger = logging.getLogger('cjn_PiTools.PiBlinky')

class PiBlinky:
    """
## Class PiBlinky (LED_name, api, gpio_num, queue, inverted=False) - A threaded, multiple LED driver for Raspberry Pi

Create an PiBlinky device instance

### Args
`LED_name` (str)
- User defined name for this LED instance, e.g., 'BLU'

`api` (str or pigpio.pi instance handle)
- 'GPIO' if using the RPi.GPIO api
- As returned from `pigpio.pi()` if using the pigpio api
- If not 'GPIO' then `api` is taken as a valid pigpio.pi handle - no error checking

`gpio_num` (int)
- GPIO channel number using "Broadcom SOC channel" numbering (e.g., int 22 refers to GPIO22 which is on header pin 15)
- Checked to be a valid GPIO channel number for Raspberry Pi versions Model B+ and later (40 pin header, values 0 to 27)

`queue` (queue handle)
- As returned from `queue.Queue()`
- No error checking

`inverted` (int, default 0)
- If 0 the GPIO channel will be driven to logic high when outputting a bit '1'
- If 1 the GPIO channel will be driven to logic low when outputting a bit '1'


### Class instance variables - as passed in at instantiation
- `LED_name` (str)
- `api` ('GPIO' or pigpio.pi instance handle)
- `gpio_num` (int)
- `queue` (queue handle)
- `inverted` (int)


### Returns
- PiBlinky instance handle on success
- Raises ValueError on args errors during instantiation, where checked


### Behaviors and rules
- Commands to a LED PiBlinky instance are passed (as a list) thru a queue to the thread managing the given 
LED - one thread and queue per LED.  
  - Command list structure:

        cmd[0]: Bittime (int or float) - Period in milliseconds for each bit
            EG, <500> is 0.5s per bit
        cmd[1]: Bitstream (str) - First bit is on the left, 1=On, 0=Off
            Spaces may be used in the bitstream for readability
            The bitstream is checked to contain only '0' and '1'
            If bitstream is an empty string then the LED is unchanged and bittime is skipped
        cmd[2]: Play count (int) - Number of times to play the bitstream
            -1 will repeat forever, until another command is queued
            0 or 1 means play the bitstream once
            2 = two times, etc
        cmd[3]: Option flag (optional enum (int))
            PiBlinky.CMD_SAVE:     Save prior command for later restore, and play the new command
            PiBlinky.CMD_RESTORE:  Restore prior command (fields 0-2 are ignored)
                The restore stack is 1 deep.  Once saved, a command may be restored more than once.
            PiBlinky.CMD_EXIT:     Play the bitstream once then exit the thread

  - A new command entered into the queue will interrupt/replace any command currently being executed.
  - On any command error, PiBlinky logs a warning message and returns.  No exception is raised.

- Debug logging may be enabled in the tool script code by setting this module's logging level:

        logging.getLogger('cjn_PiTools.PiBlinky').setLevel(logging.DEBUG)
"""

    def __init__(self, LED_name, api, gpio_num, queue, inverted=0):
        global GPIO
        self.LED_name = LED_name
        self.api = api
        self.gpio_num = gpio_num
        self.queue = queue
        self.inverted = inverted

        if not isinstance(LED_name, str):
            raise ValueError (f"LED_name must be str - received <{LED_name}> - {type(LED_name)}")

        if gpio_num not in GPIO_NUMBERS:
            raise ValueError (f"<{self.LED_name}> gpio_num must be a valid GPIO Broadcom SOC channel number - received <{gpio_num}> - {type(gpio_num)}")

        if not inverted in [0, 1]:
            raise ValueError (f"<{self.LED_name}> inverted must be int 0 or 1 - received <{inverted}> - {type(inverted)}")


        if api == 'GPIO':
            piblinky_logger.debug (f"<{self.LED_name}> - Setting up PiBlinky instance using RPi.GPIO, ch# <{self.gpio_num}>")
            import RPi.GPIO as GPIO
            GPIO.setmode(GPIO.BCM)
            GPIO.setwarnings(False)
            GPIO.setup(self.gpio_num, GPIO.OUT)
            GPIO.output(self.gpio_num, self.inverted ^ 0)
        else:       # pigpio mode
            piblinky_logger.debug (f"<{self.LED_name}> - Setting up PiBlinky instance using pigpio, ch# <{self.gpio_num}>")
            import pigpio
            self.api.set_mode(self.gpio_num, pigpio.OUTPUT)
            self.api.write(self.gpio_num, self.inverted ^ 0)


    def start(self):
        self.this_thread = Thread(target=self.piblinky, daemon=True, name='piblinky_' + self.LED_name)   # daemon so that if the main thread errors this thread is auto-killed, avoiding hang
        self.this_thread.start()
        piblinky_logger.debug (f"<{self.LED_name}> thread on GPIO {self.gpio_num:>2} created")
        return self.this_thread


    def piblinky(self):
        cmd = None
        cmd_save = None
        do_exit = False
        while True:
            cmd_prior = cmd
            cmd = self.queue.get()
            piblinky_logger.debug (f"<{self.LED_name}> Command:  <{cmd}>")

            if len(cmd) == 4:
                if cmd[3] == CMD_SAVE:
                    piblinky_logger.debug (f"<{self.LED_name}> Saving prior PiBlinky command")
                    cmd_save = cmd_prior
                elif cmd[3] == CMD_RESTORE:
                    if cmd_save != None:
                        piblinky_logger.debug (f"<{self.LED_name}> Restoring prior PiBlinky command")
                        cmd = cmd_save
                    else:
                        piblinky_logger.warning (f"<{self.LED_name}> Attempted command restore with no prior save - <{cmd}> - Skipped")
                        continue
                elif cmd[3] == CMD_EXIT:
                    do_exit = True
                else:
                    piblinky_logger.warning (f"<{self.LED_name}> Invalid option flag received - <{cmd}> - Skipped")
                    continue
            if len(cmd) == 3  or  len(cmd) == 4:
                try:
                    period    = float(cmd[0]) / 1000
                    bitstream = list(cmd[1].replace(' ', ''))   # Drop any spaces in the bitstream
                    if set(bitstream) - {'0', '1'}:
                        piblinky_logger.warning (f"<{self.LED_name}> Invalid bitstream <{bitstream}> - Skipped")
                        continue
                    play_cnt    = int(cmd[2])
                except Exception as e:
                    piblinky_logger.warning (f"<{self.LED_name}> Invalid command received - <{cmd}> - Skipped")
                    continue
            else:
                piblinky_logger.warning (f"<{self.LED_name}> Invalid command received - <{cmd}> - Skipped")
                continue
            
            piblinky_logger.debug (f"<{self.LED_name}> Stream:   period {period}s, bitstream {bitstream}, play_cnt {play_cnt}")

            quit = False
            while True:
                if not self.queue.empty()  or  quit:
                    break
                for bit in bitstream:
                    if not self.queue.empty():
                        break
                    try:
                        if self.api == 'GPIO':
                            GPIO.output(self.gpio_num, self.inverted ^ int(bit))
                        else:
                            self.api.write(self.gpio_num, self.inverted ^ int(bit))
                    except Exception as e:
                        piblinky_logger.warning (f"<{self.LED_name}> Error writing <{bit}> - Skipped:\n  {type(e).__name__}: {e}")
                        quit = True
                        break
                    time.sleep(period)
                if do_exit:
                    piblinky_logger.debug (f"<{self.LED_name}> Thread exiting")
                    sys.exit()
                if play_cnt > 0:
                    play_cnt -= 1
                if play_cnt == 0:
                    break



#!/usr/bin/env python3
"""initW1busses

1. Optionally enable power to the W1 bus(es) by setting a GPIO to 1 or 0
2. Delay to allow the kernel to scan for W1 devices, and 
3. Set user write permissions on therm_bulk_read for each found bus.

Run at boot via systemd.
"""

import importlib.metadata
__version__ = importlib.metadata.version(__package__ or __name__)


#==========================================================
#
#  Chris Nelson, Copyright 2024 - 2026
#
#==========================================================

import logging
import sys
import os
import time
import stat
import subprocess
from pathlib import Path
import argparse
import RPi.GPIO as GPIO

from cjnfuncs.core          import set_toolname
from cjnfuncs.deployfiles   import deploy_files

# Configs / Constants
TOOLNAME =                  'initW1buses'
w1_buses_root_path =        Path('/sys/devices/')

set_toolname (TOOLNAME)


def cli():
    parser = argparse.ArgumentParser(description=__doc__ + __version__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-G', '--GPIO', type=int,
                        help="Optional GPIO pin number to be set to drive --HiLo before the delay time")
    parser.add_argument('-S', '--HiLo', type=int, default=1, choices=[0, 1],
                        help="Set the --GPIO pin drive state to 0 or 1 (default 1)")
    parser.add_argument('-d', '--delay', type=int, default=20,
                        help="Delay time in seconds before setting therm_bulk_read permission (default 20)")

    parser.add_argument('-s', '--status', action='store_true',
                        help=f"Display status of W1 buses initialization")
    parser.add_argument('--setup-user', action='store_true',
                        help=f"Install starter files in user space")
    parser.add_argument('-V', '--version', action='version', version=__version__,
                        help="Print version number and exit")
    args = parser.parse_args()

    # Deploy starter files
    if args.setup_user:
        logging.getLogger('cjnfuncs.deployfiles').setLevel(logging.INFO)
        deploy_files([
            { 'source': 'initW1buses.service', 'target_dir': 'USER_CONFIG_DIR', 'file_stat': 0o644},
            ]) #, overwrite=True)
        sys.exit()


    # Status
    if args.status:
        print ("--------- Service status --------------------")
        print (f"\n{subprocess.run(['systemctl', 'status', 'initW1buses.service'],  capture_output=True, text=True).stdout}")

        print ("--------- Found buses and sensors --------------------")
        buses_list = list(w1_buses_root_path.glob('w1_bus*'))
        if len(buses_list) == 0:
            print("Found no W1 busses - Aborting")
        else:
            for bus in buses_list:
                print (bus)

                tbr = bus / 'therm_bulk_read'
                print (f"    {tbr}:   <{stat.filemode(tbr.stat().st_mode)}>")

                sensors_list = list(bus.glob('28*'))
                if len(sensors_list) == 0:
                    print ("Found no DS18B20 sensors on this bus")
                else:
                    print ("    Sensors")
                for sensor in sensors_list:
                    print (f"        {sensor}")

        sys.exit()


    # Do the initialization
    if args.GPIO:
        gpio = args.GPIO
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup(gpio, GPIO.OUT)
        GPIO.output(gpio, args.HiLo)
        logging.warning(f"Set GPIO <{gpio}> to drive <{args.HiLo}>")

    logging.warning(f"Waiting {args.delay} seconds for W1 devices discovery")
    time.sleep (args.delay)

    buses_list = list(w1_buses_root_path.glob('w1_bus*'))
    if len(buses_list) == 0:
        logging.error("Found no W1 busses - Aborting")
        return 1
    
    errored = False
    for bus in buses_list:
        try:
            tbr = bus / 'therm_bulk_read'
            os.chmod (tbr, 0o666)
            logging.warning (f"Found and enabled write access to <{tbr}>")
        except Exception as e:
            logging.warning (f"Failed to enable write access to <{tbr}> - Skipping\n  {type(e).__name__}: {e}")
            errored = True
    
    if errored:
        return 1
    else:
        return 0    # Successful completion code monitored by systemd


if __name__ == '__main__':
    sys.exit(cli())



#!/usr/bin/env python3
"""PCA9548 I2C port expander library for Raspberry Pi
"""
#==========================================================
#
#  Chris Nelson, Copyright 2026
#
#==========================================================

import sys
from cjnfuncs.core              import set_toolname, setuplogging, logging, set_logging_level

TOOLNAME =  'PCA9548'
I2C_ERROR  = -256
PCA9548_ADDRS = [0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77]

PCA9548_logger = logging.getLogger('cjn_PiTools.PCA9548')

#=====================================================================================
#=====================================================================================
#  C l a s s   P C A 9 5 4 8
#=====================================================================================
#=====================================================================================

class PCA9548:
    """
## Class PCA9548 (device_name, device_addr, pi_i2c_bus_handle) - PCA9548 I2C port expander library for Raspberry Pi

Create a PCA9548 device instance

### Args
`device_name` (str)
- User defined name for this instance, e.g., 'PCA9548'

`device_addr` (int)
- I2C bus address for this instance, e.g., 0x70
- Must be in the range of 0x70 to 0x77

`pi_i2c_bus_handle` (cjn_PiTools.shared.pi_i2c instance)
- Get an instance handle in the tools script code and pass it to this device instantiation


### Class instance variables
`device_name` (str)
- As from instantiation

`device_addr` (int)
- As from instantiation

`channel_enable_bit_map` (int)
- Current mask of enabled channels (1=enabled) - value range 0x00 to 0xFF


### Returns
- Handle to the PCA9548 instance on success
- Raises ValueError if args checks fail


### Behaviors and rules
- The device is not initialized, and need not be accessible, at instantiation (the control register is not set).
- Debug logging may be enabled in the tool script code by setting this module's logging level:

        logging.getLogger('cjn_PiTools.PCA9548').setLevel(logging.DEBUG)

    """

    def __init__ (self, device_name, device_addr, pi_i2c_bus_handle):
        self.device_name =          device_name
        self.device_addr =          device_addr
        self.pi_i2c_bus_handle =    pi_i2c_bus_handle

        if not isinstance (self.device_name, str):
            raise ValueError (f"PCA9548 device.name must be str, received <{device_name}>")

        if self.device_addr not in PCA9548_ADDRS:
            raise ValueError (f"<{self.device_name}> PCA9548 device.addr must be 0x70 - 0x77, received <0x{device_addr:0>x}>")

        api = 'smbus'  if self.pi_i2c_bus_handle.api == 'smbus'  else 'pigpio'
        PCA9548_logger.debug (f"<{self.device_name}> New PCA9548 device defined at addr <0x{self.device_addr:0>2x}> using api <{api}> on i2c bus <{self.pi_i2c_bus_handle.i2c_bus_num}>")


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ c o n t r o l _ r e g
    #=====================================================================================
    #=====================================================================================

    def write_control_reg (self, write_value):
        """
## write_control_reg (write_value) - Set the control register channel enable mask to the new write_value

***PCA9548 class member function***

The `write_value` is written to the control register and saved in `channel_enable_bit_map`


### Args
`write_value` (int or str)
- Value to be applied to the control register - resolves to between 0x00 and 0xFF
  - If int value:  Used directly as the channel enable mask
  - If str '-1':  Sets the channel enable mask to 0x00 (no channels enabled)
  - If str '0' to '7': Enables the individual channel, and disables all others, e.g., '2' resolves to 0b00000100
- The resolved value is also saved to the instance variable `channel_enable_bit_map`


### Returns
- `channel_enable_bit_map` value on success
- I2C_ERROR if unable to write to the control register
- Raises ValueError if `write_value` is not valid
        """

        PCA9548_logger.debug (f"<{self.device_name}> ***** write_control_reg()")

        channel_enable_bit_map = self._build_bit_map(write_value)

        if PCA9548_logger.isEnabledFor(logging.DEBUG):
            PCA9548_logger.debug (f"<{self.device_name}> New mask:     <0b{channel_enable_bit_map:0>8b}>, channels <{self._bit_map_to_channel_str(channel_enable_bit_map)}>")

        try:
            result = self.pi_i2c_bus_handle.i2c_write_byte(self.device_addr, channel_enable_bit_map)
            self.channel_enable_bit_map = channel_enable_bit_map
        except Exception as e:
            PCA9548_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR

        return channel_enable_bit_map


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ c o n t r o l _ r e g
    #=====================================================================================
    #=====================================================================================

    def read_control_reg (self):
        """
## read_control_reg () - Get the current channel enable mask from the control register

***PCA9548 class member function***


### Returns
- Current control register value on success
- I2C_ERROR if unable to read from the control register
        """

        PCA9548_logger.debug (f"<{self.device_name}> ***** read_control_reg()")

        try:
            current_ch_enable_mask = self.pi_i2c_bus_handle.i2c_read_byte(self.device_addr)
        except Exception as e:
            PCA9548_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR

        if PCA9548_logger.isEnabledFor(logging.DEBUG)  and  current_ch_enable_mask >= 0x00:  # TODO pigpio returns negative error codes.  Log and return I2C_ERROR
            PCA9548_logger.debug (f"<{self.device_name}> Current mask: <0b{current_ch_enable_mask:0>8b}>, channels <{self._bit_map_to_channel_str(current_ch_enable_mask)}>")
        return current_ch_enable_mask



    #=====================================================================================
    #=====================================================================================
    #  private functions
    #=====================================================================================
    #=====================================================================================

    def _build_bit_map (self, channel_value):
        """
        Accepts channel_value
            int 0x00 to 0xFF:  Set channels per bit mask
            str -1:  Unset all channels (same as int 0x00)
            str 0-7:  Set individual channel

        Returns int value in range 0x00 to 0xFF
        Raises ValueError if channel_value cannot be converted to int between -1 to 7
        """

        if isinstance(channel_value, int):
            if channel_value < 0x00  or  channel_value > 0xFF:
                raise ValueError (f"<{self.device_name}> channel_value (int) must be int range 0x00 to 0xFF - received <{channel_value}> / <0x{channel_value:0>2x}>")
            return channel_value
        
        if isinstance(channel_value, str):  # Raises ValueError if not a valid int
            xx = int(channel_value)

            if xx < -1  or  xx > 7:
                raise ValueError (f"<{self.device_name}> channel_value (str) must be int range -1 to 7 - received <{channel_value}>")
            
            if xx == -1:
                return 0x00
            else:
                return 0x01 << int(xx)
            

    def _bit_map_to_channel_str(self, bit_mask):
        """
        returns str of enabled channels for logging
        e.g., 0b00000101 returns '0 2'
        """
        ch_str = ''
        bit_mask_str = f"{bit_mask:0>8b}"
        for x in range(0, 8):
            if bit_mask_str[7-x] == '1':
                ch_str += str(x) + ' '
        return ch_str[:-1]     # trim trailing space



#=====================================================================================
#=====================================================================================
#  c l i
#=====================================================================================
#=====================================================================================

def cli():
    """PCA9548 set/get channel setting for Raspberry Pi

Mask values for set command
    0-7 enables that specific channel
    -1 sets the control register to 0b00000000 (no channels enabled)
    0xNN or 0bNNNNNNN sets the control register to this specific bit_mask (range 0x00 to 0xFF)
"""

    # import time
    import argparse
    # import datetime
    import sys
    import importlib.metadata
    __version__ = importlib.metadata.version(__package__ or __name__)

    from cjn_PiTools.shared import pi_i2c


    DEFAULT_NAME = "PCA9548"
    PCA9548_ADDRS_STR = ['0x70', '0x71', '0x72', '0x73', '0x74', '0x75', '0x76', '0x77']


    parser = argparse.ArgumentParser(description=cli.__doc__ + __version__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('Command', choices=['set', 'get'],
                        help=f"Operation command:  set or get")
    parser.add_argument('Address', choices=PCA9548_ADDRS_STR,
                        help=f"I2C address of PCA9548 device 0x70 - 0x77")
    parser.add_argument('Mask', nargs='?',
                        help=f"Set control register to this value (required for set operation)")

    parser.add_argument('-n', '--name', default=DEFAULT_NAME,
                        help=f"PCA9548 device/function name (default {DEFAULT_NAME})")
    parser.add_argument('-A', '--api', choices=['smbus', 'pigpio'], default='pigpio',
                        help=f"Either 'smbus' or 'pigpio' (default 'pigpio')")
    parser.add_argument('-H', '--host', default='localhost',
                        help=f"pigpio api target host (default 'localhost')")
    parser.add_argument('-p', '--port', type=int, default=8888,
                        help=f"pigpio api target host port number (default 8888)")

    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="Print debug-level status and activity messages")
    parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __version__,
                        help="Print version number and exit")
    args = parser.parse_args()

    set_toolname(TOOLNAME)
    setuplogging(ConsoleLogFormat="{module:>35}.{funcName:30} - {levelname:>8}:  {message}")
    set_logging_level(logging.DEBUG)

    PCA9548_logger.setLevel(logging.DEBUG)
    logging.getLogger('cjn_PiTools.shared').setLevel([logging.WARNING, logging.INFO, logging.DEBUG][args.verbose])


    # Set up interface/api
    address = int (args.Address, 16)

    if args.api == 'pigpio':
        import pigpio
        api =           pigpio.pi(args.host, args.port)
    else:
        api =           'smbus'

    pi_i2c_bus_handle = pi_i2c(api)
    PCA9548_instance =  PCA9548(args.name, address, pi_i2c_bus_handle)


    # Commands
    if args.Command == 'set':
        if args.Mask is None:
            logging.error (f"Mask value required for set command - Aborting")
            sys.exit(1)
        try:
            mask = args.Mask
            if args.Mask.startswith('0x'):
                mask = int (mask, 16)
            elif args.Mask.startswith('0b'):
                mask = int (mask, 2)
            if isinstance(mask, int)  and  (mask < 0x00  or  mask > 0xFF):
                raise ValueError ("Hex or binary value out of range 0x00 to 0xFF")
        except Exception as e:
            logging.error (f"Illegal Mask value - received <{args.Mask}> - Aborting\n  {type(e).__name__}: {e}")
            sys.exit(1)
        
        PCA9548_instance.write_control_reg(mask)


    if args.Command == 'get':
        reg_value = PCA9548_instance.read_control_reg()


    # Cleanup
    pi_i2c_bus_handle.close()
    if args.api == 'pigpio':
        api.stop()


if __name__ == '__main__':
    sys.exit(cli())



#!/usr/bin/env python3
"""ADC121C* 12-bit ADC library for Raspberry Pi
"""

#==========================================================
#
#  Chris Nelson, Copyright 2026
#
#==========================================================

import logging
from .shared import I2C_ERROR


# Configs / Constants
CONVERSION_RSLT_REG_PTR =       0x00
ALERT_STATUS_REG_PTR =          0X01
CONFIG_REG_PTR =                0x02
LOW_LIMIT_REG_PTR =             0x03
HIGH_LIMIT_REG_PTR =            0x04
HYSTERESIS_REG_PTR =            0x05
LOWEST_CONVERSION_REG_PTR =     0x06
HIGHEST_CONVERSION_REG_PTR =    0x07

ADC121_ADDRS =                  [0x50, 0x51, 0x52, 0x54, 0x55, 0x56, 0x58, 0x59, 0x5a]

adc121c_logger =                logging.getLogger('cjn_PiTools.ADC121C')


# ADC121C I2C operation:
#     The first data byte of every write operation is stored in the address pointer register.
#     This value selects the register that the following data byte(s) will be written to or read from.
#     Write a register send sequence:
#       1) ADC address, write Address Pointer Register byte, write 1 or 2 bytes of target register data (same sequence)
#     Read a register send sequence:
#       1) ADC address, write Address Pointer Register byte
#       2) ADC address, read 1 or 2 byte register data
#       If the Address Pointer is preset to the desired register, a read operation can occur without first writing 
#       the Address Pointer Register (skip step 1)


#=====================================================================================
#=====================================================================================
#  C l a s s   A D C 1 2 1 C
#=====================================================================================
#=====================================================================================

class ADC121C:
    """
## Class ADC121C (device_name, device_addr, pi_i2c_bus_handle, Vref, config_byte=None, cycle_time=0b000, alert_hold=0, alert_flag_en=0, alert_pin_en=0, polarity=0) - ADC121C* 12-bit ADC library for Raspberry Pi

Create an ADC121C family device instance

### Args
`device_name` (str)
- User defined name for this instance, e.g., 'My_ADC121C'

`device_addr` (int)
- I2C bus address for this instance, e.g., 0x50
- Allowed addresses: [0x50, 0x51, 0x52, 0x54, 0x55, 0x56, 0x58, 0x59, 0x5a]

`pi_i2c_bus_handle` (cjn_PiTools.shared.pi_i2c instance)
- Get a `pi_i2c` instance handle in the tools script code and pass it to this device instantiation

`Vref` (float)
- ADC reference (and supply) voltage
- `read_conversion_result()`'s returned 12-bit code is scaled by this value to return the measured voltage

`config_byte` (int, default None)
- If `config_byte` is an int:
  - Explicitly set the configuration register to the `config_byte` value during instantiation
  - Must be in range 0x00 to 0xFF
  - Bit 1 (Reserved) is always forced to `0`, per the specification
  - The field values within the `config_byte` are saved to the respective class instance variables, and used as the default
  values in later `write_config()` calls
  - `config_byte` takes precedent over the following individual field settings, if both are given at instantiation
- If `config_byte` is None then the following individual field settings are used

`cycle_time` (int, default 0b000)
- Value for the 3-bit Cycle Time field in the configuration register
- Must be in range 0b000 to 0b111
- Default 0b000 is Normal Mode (automatic conversion mode disabled)
- This value (or the field bits within `config_byte`, if specified) is used as the default value in later `write_config()` calls

`alert_hold` (int, default 0)
- Value for the Alert Hold field in the configuration register
- Must be 0 or 1
- Default 0 is Alert Hold disabled - alerts will self-clear
- This value (or the field bit within `config_byte`, if specified) is used as the default value in later `write_config()` calls

`alert_flag_en` (int, default 0)
- Value for the Alert Flag Enable field in the configuration register
- Must be 0 or 1
- Default 0 is Alert Flag disabled - disable alert status bit [D15] in the Conversion Result register
- This value (or the field bit within `config_byte`, if specified) is used as the default value in later `write_config()` calls

`alert_pin_en` (int, default 0)
- Value for the Alert Pin Enable field in the configuration register
- Must be 0 or 1
- Default 0 is Alert Pin disabled - disable the ALERT output pin
- This value (or the field bit within `config_byte`, if specified) is used as the default value in later `write_config()` calls
- NOTE:  The Alert Pin function is not tested - I'm using the ADC121C027 (no alert pin)

`polarity` (int, default 0)
- Value for the alert pin Polarity field in the configuration register
- Must be 0 or 1
- Default 0 - the ALERT pin is active low
- This value (or the field bit within `config_byte`, if specified) is used as the default value in later `write_config()` calls
- NOTE:  The Alert Pin function is not tested. (I'm using the ADC121C027 (no alert pin))


### Returns
- Handle to the ADC121C instance on success
- Raises ValueError if args checks fail


### Class instance variables - as passed in at instantiation
- `device_name` (str)
- `device_addr` (int)
- `Vref` (float)
- `cycle_time` (int)
- `alert_hold` (int)
- `alert_flag_en` (int)
- `alert_pin_en` (int)
- `polarity` (int)


### Behaviors and rules
- All configuration register fields default to `0`:
  - `cycle_time` defaults to 0b000 - Normal mode (automatic conversion mode disabled)
  - `alert_hold`, `alert_flag_en`, and `alert_pin_en` each default to 0 - Disabled
  - `polarity` defaults to 0 - the alert pin is active low if `alert_pin_en` is set to 1
- Debug logging may be enabled in the tool script code by setting this module's logging level:

        logging.getLogger('cjn_PiTools.ADC121C').setLevel(logging.DEBUG)
"""

    def __init__(self, device_name, device_addr, pi_i2c_bus_handle, Vref, config_byte=None,
                 cycle_time=0b000, alert_hold=0, alert_flag_en=0, alert_pin_en=0, polarity=0):

        self.device_name =          device_name
        self.device_addr =          device_addr
        self.pi_i2c_bus_handle =    pi_i2c_bus_handle
        self.Vref =                 Vref
        self.cycle_time =           cycle_time
        self.alert_hold =           alert_hold
        self.alert_flag_en =        alert_flag_en
        self.alert_pin_en =         alert_pin_en
        self.polarity =             polarity

        if not isinstance(device_name, str):
            raise ValueError (f"ADC121C device_name must be str - received <{device_name}>")

        if self.device_addr not in ADC121_ADDRS:
            xx = ", ".join(f"0x{v:02x}" for v in ADC121_ADDRS)
            yy = f"0x{device_addr:0>2x}"  if isinstance(device_addr, int)  else device_addr
            raise ValueError (f"<{self.device_name}> device address must be one of <{xx}> - received <{yy}>")

        if config_byte:
            if  not isinstance(config_byte, int)  or  config_byte < 0x00  or  config_byte > 0xFF:
                raise ValueError (f"<{self.device_name}> config_byte must be int in range 0x00 to 0xFF - received <{config_byte}>")
            self.cycle_time, self.alert_hold, self.alert_flag_en, self.alert_pin_en, self.polarity = decode_config_byte(config_byte)

        self.conv_rslt_reg_addressed = False    # Flag indicating that the conv_rslt_reg was last accessed, so skip setting the address pointer

        ADC121C.write_config(self,                              # Use base class implementation when this class is inherited
                             config_byte=   None,
                             cycle_time=    self.cycle_time,
                             alert_hold=    self.alert_hold,
                             alert_flag_en= self.alert_flag_en,
                             alert_pin_en=  self.alert_pin_en,
                             polarity=      self.polarity)

        api = 'smbus'  if self.pi_i2c_bus_handle.api == 'smbus'  else 'pigpio'
        adc121c_logger.debug (f"<{self.device_name}> New ADC121C device defined at addr <0x{self.device_addr:0>2x}> using api <{api}> on i2c bus <{self.pi_i2c_bus_handle.i2c_bus_num}>")


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ c o n v e r s i o n _ r e s u l t
    #=====================================================================================
    #=====================================================================================

    def read_conversion_result(self):
        """
## read_conversion_result () - Return the measured voltage

***ADC121C class member function***


### Returns
- Tuple (Alert Flag bit, Measured Voltage (float))
- Tuple (I2C_ERROR, I2C_ERROR) on I2C IO error

### Behaviors and rules
- The 12-bit code read from the device is scaled by Vref and returned as the measured voltage
- Setting the register Address Pointer to the Conversion Result register is skipped on consecutive calls
to `read_conversion_result()`
"""
        adc121c_logger.debug (f"<{self.device_name}> ***** read_conversion_result()")

        try:
            if not self.conv_rslt_reg_addressed:
                self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [CONVERSION_RSLT_REG_PTR])
                self.conv_rslt_reg_addressed = True
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 2)
            conv_result = (((data[0] & 0x0f) << 8) | (data[1] & 0xff)) / 4096 * self.Vref
            alert_flag = 1  if data[0] & 0x80  else 0
            adc121c_logger.debug (f"Conversion result <{self.device_name}>:  (0x{data[0]:0>2x} 0x{data[1]:0>2x}) = <{conv_result:5.3f}V>, Alert flag <{alert_flag}>")
            return alert_flag, conv_result

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR, I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ a l e r t _ s t a t u s
    #=====================================================================================
    #=====================================================================================

    def read_alert_status(self):
        """
## read_alert_status () - Return Over-range and Under-range alerts 

***ADC121C class member function***


### Returns
- Tuple (over_range_alert, under_range_alert) as integers (0 or 1)
- 1 indicates respective over-range or under-range alert
- Tuple (I2C_ERROR, I2C_ERROR) on I2C IO error
"""
        # returns tuple (over_range_alert, under_range_alert)
        adc121c_logger.debug (f"<{self.device_name}> ***** read_alert_status()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [ALERT_STATUS_REG_PTR])
            self.conv_rslt_reg_addressed = False
            alert_status_reg_value = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 1)[1][0]

            over_range_alert =      (alert_status_reg_value & 0b00000010) >> 1
            under_range_alert =     alert_status_reg_value & 0b00000001
            adc121c_logger.debug (f"over_range_alert: <{over_range_alert}>, under_range_alert: <{under_range_alert}>")
            return over_range_alert, under_range_alert

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR, I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ a l e r t _ s t a t u s
    #=====================================================================================
    #=====================================================================================

    def write_alert_status(self, clear_over=0, clear_under=0):
        """
## write_alert_status (clear_over=0, clear_under=0) - Selectively clear the Over-range and Under-range alert flags

***ADC121C class member function***


### Args
`clear_over` (int, default 0)
- Must be 0 or 1
- If 1, clear the Over Range Alert Flag

`clear_under` (int, default 0)
- Must be 0 or 1
- If 1, clear the Under Range Alert Flag


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if args checks fail
"""
        if clear_over not in [0, 1]  or  clear_under not in [0, 1]:
            raise ValueError (f"<{self.device_name}> clear_over and clear_under must be ints 0 or 1 - received <{clear_over}, {clear_under}>")

        try:
            write_byte = clear_over << 1  |  clear_under
            adc121c_logger.debug (f"<{self.device_name}> ***** write_alert_status() <0b{write_byte:0>8b}>")
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [ALERT_STATUS_REG_PTR, write_byte])
            self.conv_rslt_reg_addressed = False
            return 0

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ c o n f i g
    #=====================================================================================
    #=====================================================================================

    def write_config(self, config_byte=None, cycle_time=None, alert_hold=None, alert_flag_en=None, alert_pin_en=None, polarity=None):
        """
## write_config (config_byte=None, cycle_time=None, alert_hold=None, alert_flag_en=None, alert_pin_en=None, polarity=None) - Set operating modes

***ADC121C class member function***

The config register may be written entirely using `config_byte` or by individual field settings,
with their default values set at instantiation


### Args
`config_byte` (int, default None)
- If `config_byte` is an int:
  - Explicitly set the configuration register to the `config_byte` value
  - Must be in range 0x00 to 0xFF
  - Bit 1 (Reserved) is always forced to `0`, per the specification
  - The entire configuration register byte is written using this value, and the following args are ignored
- If `config_byte` is None then the following individual field settings are used

`cycle_time` (int, default None)
- Value for the 3-bit Cycle Time field in the configuration register
- If an int (and `config_byte` is None) then the 3-bit Cycle Time field is set to this value - must be in range 0b000 to 0b111
- If None (and `config_byte` is None), then the default value set at instantiation is used

`alert_hold` (int, default None)
- Value for the Alert Hold field in the configuration register
- If an int (and `config_byte` is None) then the field is set to this value - must be 0 or 1
- If None (and `config_byte` is None), then the default value set at instantiation is used

`alert_flag_en` (int, default None)
- Value for the Alert Flag Enable field in the configuration register
- If an int (and `config_byte` is None) then the field is set to this value - must be 0 or 1
- If None (and `config_byte` is None), then the default value set at instantiation is used

`alert_pin_en` (int, default None)
- Value for the Alert Pin Enable field in the configuration register
- If an int (and `config_byte` is None) then the field is set to this value - must be 0 or 1
- If None (and `config_byte` is None), then the default value set at instantiation is used

`polarity` (int, default None)
- Value for the Polarity field in the configuration register
- If an int (and `config_byte` is None) then the field is set to this value - must be 0 or 1
- If None (and `config_byte` is None), then the default value set at instantiation is used


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if args checks fail


### Behaviors and rules
- If `config_byte` is an int (not None), then the configuration register will be written with this value.  This 
feature gives the tool script explicit control over the configuration register.
- If `config_byte` is None then the individual field args passed in on this call are overlaid on the field value defaults established at instantiation
to construct the configuration byte.  For example:
  - At instantiation, `cycle_time=0b100`, `alert_flag_en=1`, `alert_pin_en=1` and other fields default to 0
  - On this call, `cycle_time=0b001`, `alert_hold=1`, `alert_pin_en=0`, and `polarity=1`
  - The resultant configuration byte is 0b00111001 - `cycle_time=0b001`, `alert_hold=1`, `alert_flag_en=1`, `alert_pin_en=0`, and `polarity=1`
"""

        if config_byte:
            if not isinstance(config_byte, int)  or  config_byte < 0x00  or config_byte > 0xff:
                raise ValueError (f"<{self.device_name}> config_byte must be int between 0x00 and 0xff - received <{config_byte}>")
            config_reg_value = config_byte & 0b11111101     # Bit 1 must be 0
        else:
            config_reg_value = 0x00

            _cycle_time = cycle_time  if cycle_time is not None  else self.cycle_time
            if not isinstance(_cycle_time, int)  or  _cycle_time < 0b000  or _cycle_time > 0b111:
                raise ValueError (f"<{self.device_name}> cycle_time must be int between 0b000 and 0b111 - received <{_cycle_time}>")
            config_reg_value |= _cycle_time << 5

            _alert_hold = alert_hold  if alert_hold is not None  else self.alert_hold
            if _alert_hold not in [0, 1]:
                raise ValueError (f"<{self.device_name}> alert_hold must be int 0 or 1 - received <{_alert_hold}>")
            config_reg_value |= _alert_hold << 4

            _alert_flag_en = alert_flag_en  if alert_flag_en is not None  else self.alert_flag_en
            if _alert_flag_en not in [0, 1]:
                raise ValueError (f"<{self.device_name}> alert_flag_en must be int 0 or 1 - received <{_alert_flag_en}>")
            config_reg_value |= _alert_flag_en << 3

            _alert_pin_en = alert_pin_en  if alert_pin_en is not None  else self.alert_pin_en
            if _alert_pin_en not in [0, 1]:
                raise ValueError (f"<{self.device_name}> alert_pin_en must be int 0 or 1 - received <{_alert_pin_en}>")
            config_reg_value |= _alert_pin_en << 2

            _polarity = polarity  if polarity is not None  else self.polarity
            if _polarity not in [0, 1]:
                raise ValueError (f"<{self.device_name}> polarity must be int 0 or 1 - received <{_polarity}>")
            config_reg_value |= _polarity

        adc121c_logger.debug (f"<{self.device_name}> ***** write_config() <0b{config_reg_value:0>8b}>")


        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [CONFIG_REG_PTR, config_reg_value])
            self.conv_rslt_reg_addressed = False
        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR

        return 0


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ c o n f i g
    #=====================================================================================
    #=====================================================================================

    def read_config(self):
        """
## read_config () - Return the contents of the configuration register 

***ADC121C class member function***


### Returns
- Configuration register byte on success
- I2C_ERROR on I2C IO error



### Behaviors and rules
- If debug logging is enabled for this module then the configuration register value is decoded, e.g.,

        ADC121C.read_config     -    DEBUG:  <ADC_xx> configuration register <0b00111000> settings:
        cycle_time:     <0b001>
        alert_hold:     <1>
        alert_flag_en:  <1>
        alert_pin_en:   <0>
        polarity:       <0>
"""
        # Returns config reg byte
        adc121c_logger.debug (f"<{self.device_name}> ***** read_config()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [CONFIG_REG_PTR])
            self.conv_rslt_reg_addressed = False
            config_reg_value = self.pi_i2c_bus_handle.i2c_read_device (self.device_addr, 1)[1][0]
                # i2c_read_device returns tuple (num_returned_bytes (1), [one_byte])
                # get the one_byte, which is config register value
        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR

        if adc121c_logger.isEnabledFor(logging.DEBUG):
            ct, ah, af_en, ap_en, pol = decode_config_byte(config_reg_value)

            xx = f"<{self.device_name}> configuration register <0b{config_reg_value:0>8b}> settings:\n"
            xx += f"  cycle_time:     <0b{ct:0>3b}>\n"
            xx += f"  alert_hold:     <{ah}>\n"
            xx += f"  alert_flag_en:  <{af_en}>\n"
            xx += f"  alert_pin_en:   <{ap_en}>\n"
            xx += f"  polarity:       <{pol}>"
            adc121c_logger.debug (xx)

        return config_reg_value


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ v l o w _ a l e r t _ l i m i t
    #=====================================================================================
    #=====================================================================================

    def write_vlow_alert_limit(self, vlow):
        """
## write_vlow_alert_limit (vlow) - Set the Vlow alert register voltage level

***ADC121C class member function***


### Arg
`vlow` (float or int)
- Allowed range 0 to Vref (no 'V' units)


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if arg checks fail
"""
        # returns 0 on success, or I2C_ERROR

        if not isinstance(vlow, (int, float))  or  vlow < 0.0  or  vlow > self.Vref:
            raise ValueError (f"<{self.device_name}> vlow out of range 0.0V to VRef ({self.Vref:5.3f}) - received <{vlow}>")
        vlow_reg_value = int(vlow / self.Vref * 4096)
        msB = (vlow_reg_value & 0x0F00) >> 8
        lsB = (vlow_reg_value & 0x00FF)
        try:
            adc121c_logger.debug (f"<{self.device_name}> ***** write_vlow_alert_limit() <{vlow:5.3f}V>  <[ 0x{msB:0>2x} 0x{lsB:0>2x} ]>")
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [LOW_LIMIT_REG_PTR, msB, lsB])
            self.conv_rslt_reg_addressed = False
            return 0

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ v l o w _ a l e r t _ l i m i t
    #=====================================================================================
    #=====================================================================================

    def read_vlow_alert_limit(self):
        """
## read_vlow_alert_limit () - Return the Vlow alert register voltage level

***ADC121C class member function***

The register value is converted to a voltage based on Vref


### Returns
- Vlow alert level voltage (float) on success
- I2C_ERROR on I2C IO error
"""
        adc121c_logger.debug (f"<{self.device_name}> ***** read_vlow_alert_limit()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [LOW_LIMIT_REG_PTR])
            self.conv_rslt_reg_addressed = False
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 2)
            vlow_limit_value = (((data[0] & 0x0f) << 8) | (data[1] & 0xff)) / 4096 * self.Vref
            adc121c_logger.debug (f"VLow limit <{self.device_name}>:  (0x{data[0]:0>2x} 0x{data[1]:0>2x}) = <{vlow_limit_value:5.3f}V>")
            return vlow_limit_value

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ v h i g h _ a l e r t _ l i m i t
    #=====================================================================================
    #=====================================================================================

    def write_vhigh_alert_limit(self, vhigh):
        """
## write_vhigh_alert_limit (vhigh) - Set the Vhigh alert register voltage level

***ADC121C class member function***


### Arg
`vhigh` (float or int)
- Allowed range 0 to Vref (no 'V' units)


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if arg checks fail
"""
        if not isinstance(vhigh, (int, float))  or  vhigh < 0.0  or  vhigh > self.Vref:
            raise ValueError (f"<{self.device_name}> vhigh out of range 0.0V to VRef ({self.Vref:5.3f}) - received <{vhigh}>")
        vhigh_reg_value = int((vhigh - 0.001) / self.Vref * 4096)   # Avoid vhigh=Vref resulting in 0x0000
        msB = (vhigh_reg_value & 0x0F00) >> 8
        lsB = (vhigh_reg_value & 0x00FF)
        try:
            adc121c_logger.debug (f"<{self.device_name}> ***** write_vhigh_alert_limit() <{vhigh:5.3f}V>  <[ 0x{msB:0>2x} 0x{lsB:0>2x} ]>")
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [HIGH_LIMIT_REG_PTR, msB, lsB])
            self.conv_rslt_reg_addressed = False
            return 0

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ v h i g h _ a l e r t _ l i m i t
    #=====================================================================================
    #=====================================================================================

    def read_vhigh_alert_limit(self):
        """
## read_vhigh_alert_limit () - Return the Vhigh alert register voltage level

***ADC121C class member function***

The register value is converted to a voltage based on Vref


### Returns
- Vhigh alert level voltage (float) on success
- I2C_ERROR on I2C IO error
"""
        adc121c_logger.debug (f"<{self.device_name}> ***** read_vhigh_alert_limit()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [HIGH_LIMIT_REG_PTR])
            self.conv_rslt_reg_addressed = False
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 2)
            vhigh_limit_value = (((data[0] & 0x0f) << 8) | (data[1] & 0xff)) / 4096 * self.Vref
            adc121c_logger.debug (f"VHigh limit <{self.device_name}>:  (0x{data[0]:0>2x} 0x{data[1]:0>2x}) = <{vhigh_limit_value:5.3f}V>")
            return vhigh_limit_value

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ a l e r t _ h y s t e r e s i s
    #=====================================================================================
    #=====================================================================================

    def write_alert_hysteresis(self, vhyst):
        """
## write_alert_hysteresis (vhyst) - Set the Vhyst alert register voltage level

***ADC121C class member function***


### Arg
`vhyst` (float or int)
- Allowed range 0 to Vref (no 'V' units)


### Returns
- 0 on success
- I2C_ERROR on I2C IO error
- Raises ValueError if arg checks fail
"""
        if not isinstance(vhyst, (int, float))  or  vhyst < 0.0  or  vhyst > self.Vref:
            raise ValueError (f"<{self.device_name}> vhyst out of range 0.0V to VRef ({self.Vref:5.3f}) - received <{vhyst}>")
        vhyst_reg_value = int(vhyst / self.Vref * 4096)
        msB = (vhyst_reg_value & 0x0F00) >> 8
        lsB = (vhyst_reg_value & 0x00FF)
        try:
            adc121c_logger.debug (f"<{self.device_name}> ***** write_alert_hysteresis() <{vhyst:5.3f}V>  <[ 0x{msB:0>2x} 0x{lsB:0>2x} ]>")
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [HYSTERESIS_REG_PTR, msB, lsB])
            self.conv_rslt_reg_addressed = False
            return 0

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ a l e r t _ h y s t e r e s i s
    #=====================================================================================
    #=====================================================================================

    def read_alert_hysteresis(self):
        """
## read_alert_hysteresis () - Return the Vhyst alert register voltage level

***ADC121C class member function***

The register value is converted to a voltage based on Vref


### Returns
- Vhyst alert level voltage on success
- I2C_ERROR on I2C IO error
"""
        adc121c_logger.debug (f"<{self.device_name}> ***** read_alert_hysteresis()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [HYSTERESIS_REG_PTR])
            self.conv_rslt_reg_addressed = False
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 2)
            vhyst_value = (((data[0] & 0x0f) << 8) | (data[1] & 0xff)) / 4096 * self.Vref
            adc121c_logger.debug (f"vhyst value <{self.device_name}>:  (0x{data[0]:0>2x} 0x{data[1]:0>2x}) = <{vhyst_value:5.3f}V>")
            return vhyst_value

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ l o w e s t _ c o n v e r s i o n
    #=====================================================================================
    #=====================================================================================

    def write_lowest_conversion(self):
        """
## write_lowest_conversion () - Reset the lowest conversion capture register

***ADC121C class member function***

The reset state is 0x0FFF (maximum count value, effectively Vref).


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- See notes on `read_lowest_conversion()`
"""
        msB = 0x0f          # It doesn't seem to matter what value is written to the register.  A write results in register value 0x0FFF.
        lsB = 0xff
        try:
            adc121c_logger.debug (f"<{self.device_name}> ***** write_lowest_conversion() <{self.Vref:5.3f}V>  <[ 0x{msB:0>2x} 0x{lsB:0>2x} ]>")
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [LOWEST_CONVERSION_REG_PTR, msB, lsB])
            self.conv_rslt_reg_addressed = False
            return 0

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ l o w e s t _ c o n v e r s i o n
    #=====================================================================================
    #=====================================================================================

    def read_lowest_conversion(self):
        """
## read_lowest_conversion () - Return the lowest conversion register voltage level

***ADC121C class member function***

The register value is converted to a voltage based on Vref


### Returns
- The lowest measured/captured voltage (float) on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- The datasheet says that the capture lowest/highest conversion results only works with Automatic Conversion modes (Cycle Time codes 0b001 thru 0b111).
I find that the capture lowest/highest feature also works in Normal mode (Cycle Time code 0b000).
- If switching from an Automatic Conversion mode to Normal mode note that the lowest conversion register may capture code 0x0000 (0.0V).
If using the capture lowest/highest conversion feature it is best to read these registers before stopping the Automatic Conversion
mode (switching to Normal mode).
"""
        adc121c_logger.debug (f"<{self.device_name}> ***** read_lowest_conversion()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [LOWEST_CONVERSION_REG_PTR])
            self.conv_rslt_reg_addressed = False
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 2)
            vmin_value = (((data[0] & 0x0f) << 8) | (data[1] & 0xff)) / 4096 * self.Vref
            adc121c_logger.debug (f"vmin value <{self.device_name}>:  (0x{data[0]:0>2x} 0x{data[1]:0>2x}) = <{vmin_value:5.3f}V>")
            return vmin_value

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  w r i t e _ h i g h e s t _ c o n v e r s i o n
    #=====================================================================================
    #=====================================================================================

    def write_highest_conversion(self):
        """
## write_highest_conversion () - Reset the highest conversion capture register

***ADC121C class member function***

The reset state is 0x0000 (minimum count value, effectively 0.0V).


### Returns
- 0 on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- See notes on `read_lowest_conversion()`
"""
        msB = 0x00          # It doesn't seem to matter what value is written to the register.  A write results in register value 0x0000.
        lsB = 0x00
        try:
            adc121c_logger.debug (f"<{self.device_name}> ***** write_highest_conversion() <0.000V>  <[ 0x{msB:0>2x} 0x{lsB:0>2x} ]>")
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [HIGHEST_CONVERSION_REG_PTR, msB, lsB])
            self.conv_rslt_reg_addressed = False
            return 0

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR


    #=====================================================================================
    #=====================================================================================
    #  r e a d _ h i g h e s t _ c o n v e r s i o n
    #=====================================================================================
    #=====================================================================================

    def read_highest_conversion(self):
        """
## read_highest_conversion () - Return the highest conversion register voltage level

***ADC121C class member function***

The register value is converted to a voltage based on Vref


### Returns
- The highest measured/captured voltage (float) on success
- I2C_ERROR on I2C IO error


### Behaviors and rules
- See notes on `read_lowest_conversion()`
"""
        adc121c_logger.debug (f"<{self.device_name}> ***** read_highest_conversion()")

        try:
            self.pi_i2c_bus_handle.i2c_write_device (self.device_addr, [HIGHEST_CONVERSION_REG_PTR])
            self.conv_rslt_reg_addressed = False
            (count, data) = self.pi_i2c_bus_handle.i2c_read_device(self.device_addr, 2)
            vmax_value = (((data[0] & 0x0f) << 8) | (data[1] & 0xff)) / 4096 * self.Vref
            adc121c_logger.debug (f"vmax value <{self.device_name}>:  (0x{data[0]:0>2x} 0x{data[1]:0>2x}) = <{vmax_value:5.3f}V>")
            return vmax_value

        except Exception as e:
            adc121c_logger.debug (f"<{self.device_name}> exception:  {type(e).__name__}: {e}")
            return I2C_ERROR

def decode_config_byte (config_byte):
    ct =    config_byte >> 5
    ah =    (config_byte & 0b00010000) >> 4
    af_en = (config_byte & 0b00001000) >> 3
    ap_en = (config_byte & 0b00000100) >> 2
    pol =   config_byte & 0b00000001
    return ct, ah, af_en, ap_en, pol
